import logging
from typing import Any, Callable

from ..services.service import ResnapService
from .metadata import Metadata
from .signature import get_function_signature
from .utils import hash_arguments

logger = logging.getLogger("resnap")


class ResultsRetriever:
    def __init__(self, service: ResnapService, options: dict[str, Any]) -> None:
        self._service = service
        self._considered_attributes: list[str] = options.get("considered_attributes", [])
        self._enable_recovery: bool = options.get("enable_recovery", True)
        self._consider_args: bool = options.get("consider_args", True)

        self.output_folder: str = options.get("output_folder", "")
        self.func_name: str = ""
        self.hashed_arguments: str = ""

    def get_results(self, func: Callable, args: tuple, kwargs: dict) -> tuple[Any, bool]:
        """
        Get the results from the resnap service.

        Args:
            func (Callable): The function to get the signature from.
            args (tuple): The arguments passed to the function.
            kwargs (dict): The keyword arguments passed to the function.

        Returns:
            tuple[Any, bool]: The result of the function and a boolean indicating if the result was recovered.
        """
        self._service.create_output_folder(self.output_folder)
        self.func_name, arguments = get_function_signature(func, args, kwargs, self._considered_attributes)
        self.hashed_arguments: str = hash_arguments(arguments)
        return self._get_saved_result()

    def _get_saved_result(self) -> tuple[Any, bool]:
        if not self._enable_recovery:
            return None, False

        metadata: list[Metadata] = self._service.get_success_metadata(self.func_name, self.output_folder)
        if not metadata:
            return None, False

        if not self._consider_args:
            return self._service.read_result(metadata[0]), True

        for metadata_entry in metadata:
            if metadata_entry.hashed_arguments == self.hashed_arguments:
                logger.debug("Returning saved result...")
                return self._service.read_result(metadata_entry), True

        return None, False



from typing import Any

import pytest

from resnap.helpers.signature import get_function_signature


def function_to_test(test: str = "test") -> None:
    pass


class MockClass:
    MAX_VALUE: int = 10

    def __init__(self, min_value: int = 0) -> None:
        self._min_value = min_value

    @classmethod
    def class_method(cls, test: str = "test") -> None:
        pass

    @staticmethod
    def static_method(test: str = "test", **kwargs: Any) -> None:
        pass

    def method(self, test: str = "test", **kwargs: Any) -> None:
        pass


class TestGetFunctionSignature:
    @pytest.mark.parametrize(
        "my_args, expected_result",
        [
            (("test",), {"test": "test", "MAX_VALUE": 10}),
            (("toto",), {"test": "toto", "MAX_VALUE": 10}),
            ((), {"test": "test", "MAX_VALUE": 10}),
        ],
    )
    def test_should_return_function_signature_with_class_method(self, my_args: tuple, expected_result: dict) -> None:
        # Given
        args = (MockClass, *my_args)

        # When
        signature, arguments = get_function_signature(MockClass.class_method.__func__, args, {}, ["MAX_VALUE"])

        # Then
        assert signature == "MockClass.class_method"
        assert arguments == expected_result

    @pytest.mark.parametrize(
        "my_args, expected_result",
        [
            (("test",), {"test": "test", "MAX_VALUE": 10, "_min_value": 5}),
            (("toto",), {"test": "toto", "MAX_VALUE": 10, "_min_value": 5}),
            ((), {"test": "test", "MAX_VALUE": 10, "_min_value": 5}),
        ],
    )
    def test_should_return_function_signature_with_class_method_instance(
        self,
        my_args: tuple,
        expected_result: dict,
    ) -> None:
        # Given
        args = (MockClass(5), *my_args)

        # When
        signature, arguments = get_function_signature(
            MockClass.class_method.__func__, args,
            {},
            ["MAX_VALUE", "_min_value"],
        )

        # Then
        assert signature == "MockClass.class_method"
        assert arguments == expected_result

    @pytest.mark.parametrize(
        "my_args, my_kwargs, expected_result",
        [
            (("test",), {"titi": 1}, {"test": "test", "kwargs": {"titi": 1}}),
            (("toto",), {}, {"test": "toto", "kwargs": {}}),
            ((), {"titi": 1}, {"test": "test", "kwargs": {"titi": 1}}),
        ],
    )
    def test_should_return_function_signature_with_static_method(
        self, my_args: tuple, my_kwargs: dict, expected_result: dict
    ) -> None:
        # When
        signature, arguments = get_function_signature(MockClass.static_method, my_args, my_kwargs)

        # Then
        assert signature == "MockClass.static_method"
        assert arguments == expected_result

    def test_should_return_function_signature_with_method(self) -> None:
        # Given
        obj = MockClass()
        my_args = (obj,)
        my_kwargs = {}

        # When
        signature = get_function_signature(MockClass.method, my_args, my_kwargs, ["MAX_VALUE", "_min_value"])

        # Then
        assert signature == ("MockClass.method", {"test": "test", "kwargs": {}, "MAX_VALUE": 10, "_min_value": 0})

    def test_should_return_function_signature_with_method_without_class_attributes(self) -> None:
        # Given
        obj = MockClass()
        my_args = (obj,)
        my_kwargs = {}

        # When
        signature = get_function_signature(MockClass.method, my_args, my_kwargs)

        # Then
        assert signature == ("MockClass.method", {"test": "test", "kwargs": {}})

    def test_should_return_function_signature_with_function(self) -> None:
        # Given
        args = ()
        kwargs = {}

        # When
        signature = get_function_signature(function_to_test, args, kwargs)

        # Then
        assert signature == ("function_to_test", {"test": "test"})



from datetime import datetime
from typing import Any

import pytest

from resnap.helpers.metadata import Metadata, MetadataFail, MetadataSuccess
from resnap.helpers.status import Status
from resnap.helpers.utils import hash_arguments

hashed_arguments: str = hash_arguments({"key": "value"})


class TestMetadata:
    @pytest.mark.parametrize(
        "data, expected",
        [
            (
                {
                    "status": "SUCCESS",
                    "event_time": "2021-01-01T00:00:00",
                    "hashed_arguments": hashed_arguments,
                    "result_path": "/path/to/result",
                    "result_type": "str",
                },
                MetadataSuccess,
            ),
            (
                {
                    "status": "FAIL",
                    "event_time": "2021-01-01T00:00:00",
                    "hashed_arguments": hash_arguments({}),
                    "error_message": "Something went wrong",
                },
                MetadataFail,
            ),
        ],
    )
    def test_should_return_metadata(self, data: dict[str, Any], expected: Metadata) -> None:
        # When
        metadata = Metadata.from_dict(data)

        # Then
        assert isinstance(metadata, expected)


class TestMetadataSuccess:
    def test_should_return_dict(self) -> None:
        # Given
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hashed_arguments,
            result_path="/path/to/result",
            result_type="str",
            extra_metadata={"key": "value"},
        )

        # When
        result = metadata.to_dict()

        # Then
        assert result == {
            "status": "SUCCESS",
            "event_time": "2021-01-01T00:00:00",
            "hashed_arguments": hashed_arguments,
            "result_path": "/path/to/result",
            "result_type": "str",
            "extra_metadata": {"key": "value"},
        }

    def test_should_return_metadata_from_dict(self) -> None:
        # Given
        data = {
            "status": "SUCCESS",
            "event_time": "2021-01-01T00:00:00",
            "hashed_arguments": hashed_arguments,
            "result_path": "/path/to/result",
            "result_type": "str",
            "extra_metadata": {"key": "value"},
        }

        # When
        metadata = MetadataSuccess.from_dict(data)

        # Then
        assert metadata.status == Status.SUCCESS
        assert metadata.event_time == datetime.fromisoformat("2021-01-01T00:00:00")
        assert metadata.hashed_arguments == hashed_arguments
        assert metadata.result_path == "/path/to/result"
        assert metadata.result_type == "str"
        assert metadata.extra_metadata == {"key": "value"}


class TestMetadataFail:
    def test_should_return_dict_without_data(self) -> None:
        # Given
        metadata = MetadataFail(
            status=Status.FAIL,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hashed_arguments,
            error_message="oopsi an error",
            data={},
            extra_metadata={},
        )

        # When
        result = metadata.to_dict()

        # Then
        assert result == {
            "status": "FAIL",
            "event_time": "2021-01-01T00:00:00",
            "hashed_arguments": hashed_arguments,
            "error_message": "oopsi an error",
        }

    def test_should_return_dict_with_data(self) -> None:
        # Given
        metadata = MetadataFail(
            status=Status.FAIL,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hashed_arguments,
            error_message="oopsi an error",
            data={"key": "value"},
            extra_metadata={"key": "value"},
        )

        # When
        result = metadata.to_dict()

        # Then
        assert result == {
            "status": "FAIL",
            "event_time": "2021-01-01T00:00:00",
            "hashed_arguments": hashed_arguments,
            "error_message": "oopsi an error",
            "data": {"key": "value"},
            "extra_metadata": {"key": "value"},
        }

    def test_should_return_metadata_from_dict(self) -> None:
        # Given
        data = {
            "status": "FAIL",
            "event_time": "2021-01-01T00:00:00",
            "hashed_arguments": hashed_arguments,
            "error_message": "oopsi an error",
            "extra_metadata": {"key": "value"},
        }

        # When
        metadata = MetadataFail.from_dict(data)

        # Then
        assert metadata.status == Status.FAIL
        assert metadata.event_time == datetime.fromisoformat("2021-01-01T00:00:00")
        assert metadata.hashed_arguments == hashed_arguments
        assert metadata.error_message == "oopsi an error"
        assert not metadata.data
        assert metadata.extra_metadata == {"key": "value"}



import time
from pathlib import Path

import pandas as pd
import toml
from constants import FILES, TEST_CONFIG_GLOBAL
from job_test import run_job

TEST_CONFIG_LOCAL = TEST_CONFIG_GLOBAL.copy()
RESULT_PATH = Path(toml.load("pyproject.toml")["tool"]["resnap"]["output_base_path"])


def execution_no_previous_save() -> list[Path]:
    run_job(conf_test=TEST_CONFIG_LOCAL)

    generated_files = list(RESULT_PATH.rglob("*.resnap*"))
    for file in generated_files:
        if not [f for f in FILES if f in file.name]:
            assert False, "No file was generated"

    assert len(generated_files) == len(FILES) * 2, "Not all files were generated"
    return generated_files


def execution_with_previous_save_same_config(last_generated_files: list[Path]) -> None:
    generated_files = execution_no_previous_save()
    assert sorted(generated_files) == sorted(last_generated_files), "Don't use resnap"


def execution_with_previous_save_different_config(last_generated_files: list[Path]) -> list[Path]:
    config = TEST_CONFIG_LOCAL.copy()
    config["first_method_argument"] = pd.DataFrame(
        {
            "C": [1, 2, 3],
            "D": [4, 5, 6],
            "E": [7, 8, 9],
        }
    )
    config["second_method_argument"] = "toto"
    config["third_method_argument"] = "tata"
    config["BasicClass_get_param_value_argument"] = "toto"
    config["BasicClass_generate_dataframe_argument"] = "toto"
    config["a_function_argument"] = False
    run_job(conf_test=config)

    generated_files = list(RESULT_PATH.rglob("*.resnap*"))
    new_files = []
    for file in generated_files:
        if [f for f in FILES if f in file.name and file not in last_generated_files]:
            new_files.append(file)

    if not new_files:
        assert False, "No file was generated"

    assert len(generated_files) == len(FILES) * 4, "Not all files were generated"
    assert len(new_files) == len(FILES) * 2, "Not all files were generated"
    assert sorted(new_files) != sorted(last_generated_files), "Don't use resnap"
    return new_files


def execution_with_clean(last_generated_files: list[Path]) -> None:
    generated_files = execution_no_previous_save()
    assert sorted(generated_files) != sorted(last_generated_files), "Don't clean old files"
    assert len(generated_files) == len(FILES) * 2, "Not all files were generated"


def clear_project(path: Path = RESULT_PATH) -> None:
    if not path.exists():
        return
    for file in path.iterdir():
        if file.is_file():
            file.unlink()
        else:
            clear_project(file)
    path.rmdir()


def run() -> None:
    print()
    print("**** START TESTS LOCAL SERVICE ****")
    clear_project()

    print("Running test: first execution with no save")
    generated_files = execution_no_previous_save()
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with save and same configuration")
    execution_with_previous_save_same_config(generated_files)
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with save and different configuration")
    generated_files = execution_with_previous_save_different_config(generated_files)
    print("Test passed")
    print("-" * 25)
    time.sleep(3)

    print("Running test: execution with same configuration, must clean old files")
    execution_with_clean(generated_files)
    print("Test passed")
    print("-" * 25)

    print("Cleaning files generated by tests")
    clear_project()
    print("-" * 25)

    print("**** END TESTS LOCAL SERVICE ****")



import functools
import logging
from collections.abc import Callable, Coroutine
from datetime import datetime
from typing import Any, ParamSpec, TypeVar

from .exceptions import ResnapError
from .factory import ResnapServiceFactory
from .helpers.context import clear_metadata, get_metadata, restore_metadata
from .helpers.results_retriever import ResultsRetriever
from .services.service import ResnapService

logger = logging.getLogger("resnap")


def _save(
    service: ResnapService,
    func_name: str,
    output_folder: str,
    hashed_arguments: str,
    result: Any,
    extra_metadata: dict,
    output_format: str | None = None,
) -> None:
    logger.debug("Saving result...")
    result_path, event_time = service.save_result(func_name, result, output_folder, output_format)
    service.save_success_metadata(
        func_name=func_name,
        output_folder=output_folder,
        hashed_arguments=hashed_arguments,
        event_time=event_time,
        result_path=result_path,
        result_type=type(result).__name__,
        extra_metadata=extra_metadata,
    )


def _clear(service: ResnapService) -> None:
    logger.debug("Clearing old saves...")
    service.clear_old_saves()


R = TypeVar("R")  # Return type
P = ParamSpec("P")  # Parameter specification


def resnap(_func: Callable[P, R] | None = None, **options: Any) -> Callable[P, R]:
    """
    Decorator to save the result of a function and return it if the function has been called with the same arguments
    before and the result was saved successfully.

    Args:
        output_format (str): The format in which the result should be saved.
            If None, the result will be saved in the default format of the service.
            Allowed formats for pd.DataFrame: "parquet" (default) and "csv".
            Allowed formats for other types: "pickle" (default), "txt" and "json".
        output_folder (str): The folder where the result should be saved.
            If None, the result will be saved in the default folder of the service.
        enable_recovery (bool): If True, the recovery system is enabled. The decorator will attempt to load the last
            saved result if the input parameters are identical. If False, the backup will be performed. Default is True.
        consider_args (bool): If True, the function arguments are considered to match the last saved result.
            If False, the arguments are ignored and only the function name is considered. Default is True.
        considered_attributes (list[str]): The list of class/instance attributes to consider when hashing the function
            arguments. Warning: Do not use __slots__ in your classes if you want to use this feature.
    """
    if _func is None:
        return functools.partial(resnap, **options)

    @functools.wraps(_func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        service: ResnapService = ResnapServiceFactory.get_service()

        if not service.is_enabled:
            return _func(*args, **kwargs)
        _clear(service)
        token = clear_metadata()

        results_retriever = ResultsRetriever(service, options)
        result, is_recovery = results_retriever.get_results(_func, args, kwargs)
        if is_recovery:
            return result

        try:
            logger.debug(f"Executing function {results_retriever.func_name}...")
            result = _func(*args, **kwargs)
            _save(
                service=service,
                func_name=results_retriever.func_name,
                output_folder=results_retriever.output_folder,
                hashed_arguments=results_retriever.hashed_arguments,
                result=result,
                extra_metadata=get_metadata(),
                output_format=options.get("output_format"),
            )
            return result

        except Exception as e:
            service.save_failed_metadata(
                func_name=results_retriever.func_name,
                output_folder=results_retriever.output_folder,
                hashed_arguments=results_retriever.hashed_arguments,
                event_time=datetime.now(),
                error_message=str(e),
                data=e.data if isinstance(e, ResnapError) else {},
                extra_metadata=get_metadata(),
            )
            raise e
        finally:
            restore_metadata(token)

    return wrapper


def async_resnap(
    _func: Callable[P, Coroutine[Any, Any, R]] | None = None,
    **options: Any,
) -> Callable[P, Coroutine[Any, Any, R]]:
    """
    Decorator to save the result of an async function and return it if the function has been called
    with the same arguments before and the result was saved successfully.

    Args:
        output_format (str): The format in which the result should be saved.
            If None, the result will be saved in the default format of the service.
            Allowed formats for pd.DataFrame: "parquet" (default) and "csv".
            Allowed formats for other types: "pickle" (default), "txt" and "json".
        output_folder (str): The folder where the result should be saved.
            If None, the result will be saved in the default folder of the service.
        enable_recovery (bool): If True, the recovery system is enabled. The decorator will attempt to load the last
            saved result if the input parameters are identical. If False, the backup will be performed. Default is True.
        consider_args (bool): If True, the function arguments are considered to match the last saved result.
            If False, the arguments are ignored and only the function name is considered. Default is True.
        considered_attributes (list[str]): The list of class/instance attributes to consider when hashing the function
            arguments.
    """
    if _func is None:
        return functools.partial(async_resnap, **options)

    @functools.wraps(_func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        service: ResnapService = ResnapServiceFactory.get_service()

        if not service.is_enabled:
            return await _func(*args, **kwargs)
        _clear(service)
        token = clear_metadata()

        results_retriever = ResultsRetriever(service, options)
        result, is_recovery = results_retriever.get_results(_func, args, kwargs)
        if is_recovery:
            return result

        try:
            logger.debug(f"Executing function {results_retriever.func_name}...")
            result = await _func(*args, **kwargs)
            _save(
                service=service,
                func_name=results_retriever.func_name,
                output_folder=results_retriever.output_folder,
                hashed_arguments=results_retriever.hashed_arguments,
                result=result,
                extra_metadata=get_metadata(),
                output_format=options.get("output_format"),
            )
            return result

        except Exception as e:
            service.save_failed_metadata(
                func_name=results_retriever.func_name,
                output_folder=results_retriever.output_folder,
                hashed_arguments=results_retriever.hashed_arguments,
                event_time=datetime.now(),
                error_message=str(e),
                data=e.data if isinstance(e, ResnapError) else {},
                extra_metadata=get_metadata(),
            )
            raise e
        finally:
            restore_metadata(token)

    return wrapper



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any

import pandas as pd

from ..helpers.config import Config
from ..helpers.constants import EXT, SEPARATOR
from ..helpers.metadata import Metadata, MetadataFail, MetadataSuccess
from ..helpers.singleton import SingletonABCMeta
from ..helpers.status import Status


class ResnapService(ABC, metaclass=SingletonABCMeta):
    def __init__(self, config: Config) -> None:
        self.config: Config = config

    @property
    def is_enabled(self) -> bool:
        return self.config.enabled

    def metadata_path(self, func_name: str, event_time: datetime, output_folder: str) -> str:
        """
        Get the metadata path based on the function name and event time.

        Args:
            func_name (str): The function name.
            event_time (datetime): The event time.
            output_folder (str): The output folder.
        Returns:
            str: The metadata path.
        """
        parts = []
        if self.config.output_base_path:
            parts.append(self.config.output_base_path)
        if output_folder:
            parts.append(output_folder)
        parts.append(f"{func_name}_{event_time.isoformat().replace(':', '-')}{EXT}")
        return SEPARATOR.join(parts)

    def result_path(self, func_name: str, event_time: datetime, output_folder: str, output_ext: str) -> str:
        """
        Get the result path based on the function name, event time, and output extension.

        Args:
            func_name (str): The function name.
            event_time (datetime): The event time.
            output_folder (str): The output folder.
            output_ext (str): The output extension.
        Returns:
            str: The result path.
        """
        parts = []
        if self.config.output_base_path:
            parts.append(self.config.output_base_path)
        if output_folder:
            parts.append(output_folder)
        parts.append(f"{func_name}_{event_time.isoformat().replace(':', '-')}{EXT}.{output_ext}")
        return SEPARATOR.join(parts)

    @abstractmethod
    def clear_old_saves(self) -> None:  # pragma: no cover
        """
        Clear old saves (metadata and saved files) based on the configuration.
        """
        raise NotImplementedError

    def _get_output_path(self, output_folder: str) -> str:
        output_path = self.config.output_base_path
        if output_folder:
            output_path += f"/{output_folder}"
        return output_path

    @abstractmethod
    def _create_folder(self, path: str, folder_name: str) -> None:  # pragma: no cover
        """
        Create the output folder based on the configuration.
        Args:
            path (str): The path to create the folder.
            folder_name (str): The name of the folder.
        """
        raise NotImplementedError

    def create_output_folder(self, output_folder: str) -> None:
        """
        Create the output folder based on the configuration.

        Args:
            output_folder (str): The output folder.
        """
        if self.config.output_base_path:
            self._create_folder("", self.config.output_base_path)
        if output_folder:
            self._create_folder(self.config.output_base_path, output_folder)

    @abstractmethod
    def get_success_metadata(self, func_name: str, output_folder: str) -> list[Metadata]:  # pragma: no cover
        """
        Get all success metadata based on the function name.

        Args:
            func_name (str): The function name.
            output_folder (str): The output folder.
        Returns:
            list[Metadata]: The success metadata based on the function name.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_parquet_to_dataframe(self, file_path: str) -> pd.DataFrame:  # pragma: no cover
        """
        Read parquet file from the given file path.

        Args:
            file_path (str): The file path to read the parquet file.
        Returns:
            pd.DataFrame: The read parquet file.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_csv_to_dataframe(self, file_path: str) -> pd.DataFrame:  # pragma: no cover
        """
        Read csv file from the given file path.

        Args:
            file_path (str): The file path to read the csv file.
        Returns:
            pd.DataFrame: The read csv file.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_pickle(self, file_path: str) -> Any:  # pragma: no cover
        """
        Read pickle file from the given file path.

        Args:
            file_path (str): The file path to read the pickle file.
        Returns:
            Any: The read pickle file.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_text(self, file_path: str) -> Any:  # pragma: no cover
        """
        Read pickle file from the given file path.

        Args:
            file_path (str): The file path to read the pickle file.
        Returns:
            Any: The read pickle file.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_json(self, file_path: str) -> Any:  # pragma: no cover
        """
        Read json file from the given file path.

        Args:
            file_path (str): The file path to read the json file.
        Returns:
            Any: The read json file.
        """
        raise NotImplementedError

    def read_result(self, metadata: MetadataSuccess) -> Any:
        """
        Read the result based on the metadata.

        Args:
            metadata (MetadataSuccess): The metadata to read the result.
        Returns:
            Any: The read result.
        """
        file_path: str = metadata.result_path
        result_type: str = metadata.result_type
        need_cast: bool = False

        if "DataFrame" in result_type:
            if file_path.endswith(".gz"):
                func = self._read_parquet_to_dataframe
            elif file_path.endswith(".csv"):
                func = self._read_csv_to_dataframe
            else:
                raise NotImplementedError(f"Unsupported result type: {result_type}")
        elif file_path.endswith(".txt"):
            need_cast: bool = True
            func = self._read_text
        elif file_path.endswith(".json"):
            func = self._read_json
        elif file_path.endswith(".pkl"):
            func = self._read_pickle
        else:
            raise NotImplementedError(f"Unsupported result type: {result_type}")

        result = func(file_path)
        if need_cast:
            var_type = eval(result_type)
            result = var_type(result)
        return result

    @abstractmethod
    def _save_dataframe_to_csv(self, result: pd.DataFrame, result_path: str) -> None:  # pragma: no cover
        """
        Save dataframe to csv file from the given result and result path.

        Args:
            result (pd.DataFrame): The result.
            result_path (str): The result path.
        """
        raise NotImplementedError

    @abstractmethod
    def _save_dataframe_to_parquet(self, result: pd.DataFrame, result_path: str) -> None:  # pragma: no cover
        """
        Save dataframe to parquet file from the given result and result path.

        Args:
            result (pd.DataFrame): The result.
            result_path (str): The result path.
        """
        raise NotImplementedError

    @abstractmethod
    def _save_to_pickle(self, result: Any, result_path: str) -> None:  # pragma: no cover
        """
        Save pickle file from the given result and result path.

        Args:
            result (Any): The result.
            result_path (str): The result path.
        """
        raise NotImplementedError

    @abstractmethod
    def _save_to_text(self, result: Any, result_path: str) -> None:  # pragma: no cover
        """
        Save text file from the given result and result path.

        Args:
            result (Any): The result.
            result_path (str): The result path.
        """
        raise NotImplementedError

    @abstractmethod
    def _save_to_json(self, result: Any, result_path: str) -> None:  # pragma: no cover
        """
        Save json file from the given result and result path.

        Args:
            result (Any): The result.
            result_path (str): The result path.
        """
        raise NotImplementedError

    def save_result(
        self,
        func_name: str,
        result: Any,
        output_folder: str,
        output_format: str | None = None,
    ) -> tuple[str, datetime]:
        """
        Save the result based on the function name and result.

        Args:
            func_name (str): The function name.
            result (Any): The result to save.
            output_folder (str): The output folder.
            output_format (str | None): The output format.
        Returns:
            tuple[str, datetime]: The result path and event time.
        """
        if type(result) is pd.DataFrame:
            if output_format == "csv":
                func = self._save_dataframe_to_csv
            else:
                output_format = "parquet.gz"
                func = self._save_dataframe_to_parquet
        elif output_format == "txt":
            func = self._save_to_text
        elif output_format == "json":
            func = self._save_to_json
        else:
            output_format = "pkl"
            func = self._save_to_pickle

        event_time: datetime = datetime.now()
        result_path = self.result_path(func_name, event_time, output_folder, output_format)
        func(result=result, result_path=result_path)
        return result_path, event_time

    @abstractmethod
    def _write_metadata(self, metadata_path: str, metadata: Metadata) -> None:  # pragma: no cover
        """
        Write metadata to the given metadata path.

        Args:
            metadata_path (Path): The metadata path to write the metadata.
            metadata (Metadata): The metadata to write.
        """
        raise NotImplementedError

    def _save_metadata(self, func_name: str, metadata: Metadata, output_folder: str) -> None:
        """
        Save the metadata based on the function name and metadata.

        Args:
            func_name (str): The function name.
            metadata (Metadata): The metadata to save.
            output_folder (str): The output folder.
        """
        metadata_path: str = self.metadata_path(func_name, metadata.event_time, output_folder)
        self._write_metadata(metadata_path, metadata)

    def save_success_metadata(
        self,
        func_name: str,
        output_folder: str,
        hashed_arguments: str,
        event_time: datetime,
        result_path: str,
        result_type: str,
        extra_metadata: dict,
    ) -> None:
        """
        Save success metadata based on the function name, arguments, event time, and result path.

        Args:
            func_name (str): The function name.
            output_folder (str): The output folder.
            hashed_arguments (str): The hashed arguments.
            event_time (datetime): The event time.
            result_path (str): The result path.
            result_type (str): The result type.
            extra_metadata (dict): The extra metadata.
        """
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=event_time,
            hashed_arguments=hashed_arguments,
            result_path=result_path,
            result_type=result_type,
            extra_metadata=extra_metadata,
        )
        self._save_metadata(func_name, metadata, output_folder)

    def save_failed_metadata(
        self,
        func_name: str,
        output_folder: str,
        hashed_arguments: str,
        event_time: datetime,
        error_message: str,
        data: dict,
        extra_metadata: dict,
    ) -> None:
        """
        Save failed metadata based on the function name, arguments, event time, error message, and data.

        Args:
            func_name (str): The function name.
            output_folder (str): The output folder.
            hashed_arguments (str): The function's hashed arguments.
            event_time (datetime): The event time.
            error_message (str): The error message.
            data (dict): The data to save.
        """
        metadata = MetadataFail(
            status=Status.FAIL,
            event_time=event_time,
            hashed_arguments=hashed_arguments,
            error_message=error_message,
            data=data,
            extra_metadata=extra_metadata,
        )
        self._save_metadata(func_name, metadata, output_folder)



import os

from resnap.helpers.ssl import get_ca_bundle_path


def test_returns_env_var_if_set(monkeypatch) -> None:
    monkeypatch.setenv("REQUESTS_CA_BUNDLE", "/custom/path/to/ca.pem")
    assert get_ca_bundle_path() == "/custom/path/to/ca.pem"


def test_returns_stripped_env_var(monkeypatch) -> None:
    monkeypatch.setenv("REQUESTS_CA_BUNDLE", "   /custom/path/with/spaces.pem   ")
    assert get_ca_bundle_path() == "/custom/path/with/spaces.pem"


def test_returns_false_on_non_posix_os(monkeypatch) -> None:
    monkeypatch.delenv("REQUESTS_CA_BUNDLE", raising=False)
    monkeypatch.setattr(os, "name", "nt")
    monkeypatch.setattr(os.path, "exists", lambda path: True)
    assert get_ca_bundle_path() is False



import time

import pandas as pd
import urllib3
from constants import FILES, TEST_CONFIG_GLOBAL
from job_test import run_job

from resnap import set_resnap_service
from resnap.boto.client import S3Client
from resnap.boto.config import S3Config
from resnap.helpers.config import Config
from resnap.helpers.utils import load_file
from resnap.services.boto_service import BotoResnapService

urllib3.disable_warnings()
RESULT_PATH = "resnap"
TEST_CONFIG_CEPH = TEST_CONFIG_GLOBAL.copy()
CONFIG = Config(**{
    "enabled": True,
    "save_to": "s3",
    "output_base_path": RESULT_PATH,
    "secrets_file_name": "boto_secrets.yml",
    "enable_remove_old_files": True,
    "max_history_files_length": 60,
    "max_history_files_time_unit": "second",
})


def set_service() -> BotoResnapService:
    set_resnap_service(BotoResnapService(CONFIG))


def get_s3_client() -> S3Client:
    config = S3Config(**load_file(CONFIG.secrets_file_name, key="resnap"))
    return S3Client(config)


def get_files_without_folder() -> list[str]:
    s3_client = get_s3_client()
    files = []
    for file in s3_client.list_objects(RESULT_PATH, True):
        if not file.endswith('/'):
            files.append(file)
    return files


def execution_no_previous_save() -> list[str]:
    run_job(conf_test=TEST_CONFIG_CEPH.copy())

    generated_files = get_files_without_folder()
    for file in generated_files:
        if not [f for f in FILES if f in file]:
            assert False, "No file was generated"

    assert len(generated_files) == len(FILES) * 2, "Not all files were generated"
    return generated_files


def execution_with_previous_save_same_config(last_generated_files: list[str]) -> None:
    generated_files = execution_no_previous_save()
    assert sorted(generated_files) == sorted(last_generated_files), "Don't use checkpoint"


def execution_with_previous_save_different_config(last_generated_files: list[str]) -> list[str]:
    config = TEST_CONFIG_CEPH.copy()
    config["first_method_argument"] = pd.DataFrame(
        {
            "C": [1, 2, 3],
            "D": [4, 5, 6],
            "E": [7, 8, 9],
        }
    )
    config["second_method_argument"] = "toto"
    config["third_method_argument"] = "tata"
    config["BasicClass_get_param_value_argument"] = "toto"
    config["BasicClass_generate_dataframe_argument"] = "toto"
    config["a_function_argument"] = False
    run_job(conf_test=config)

    generated_files = get_files_without_folder()
    new_files = []
    for file in generated_files:
        if [f for f in FILES if f in file and file not in last_generated_files]:
            new_files.append(file)

    if not new_files:
        assert False, "No file was generated"

    assert len(generated_files) == len(FILES) * 4, "Not all files were generated"
    assert len(new_files) == len(FILES) * 2, "Not all files were generated"
    assert sorted(new_files) != sorted(last_generated_files), "Don't use checkpoint"
    return new_files


def execution_with_clean(last_generated_files: list[str]) -> None:
    generated_files = execution_no_previous_save()
    assert sorted(generated_files) != sorted(last_generated_files), "Don't clean old files"
    assert len(generated_files) == len(FILES) * 2, "Not all files were generated"


def clear_project() -> None:
    s3_client = get_s3_client()
    if s3_client.object_exists(f"{RESULT_PATH}/"):
        s3_client.rmdir(RESULT_PATH)


def run() -> None:
    print()
    print("**** START TESTS CEPH SERVICE ****")
    set_service()
    clear_project()

    print("Running test: execution with no save")
    generated_files = execution_no_previous_save()
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with save and same configuration")
    execution_with_previous_save_same_config(generated_files)
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with save and different configuration")
    generated_files = execution_with_previous_save_different_config(generated_files)
    print("Test passed")
    print("-" * 25)
    time.sleep(60)

    print("Running test: execution with same configuration, must clean old files")
    execution_with_clean(generated_files)
    print("Test passed")
    print("-" * 25)

    print("Cleaning files generated for tests")
    clear_project()
    print("-" * 25)

    print("**** END TESTS LOCAL SERVICE ****")



import io
import zlib
from abc import ABC, abstractmethod
from os import path

import pandas as pd
import pyarrow as pa

BufferType = io.FileIO | io.BytesIO | io.StringIO
_SUPPORTED_COMPRESSIONS = ["gzip"]
AVAILABLE_COMPRESSIONS = {"gzip": [".gz"], "snappy": ["*"]}


class DataFrameHandler(ABC):
    @staticmethod
    @abstractmethod
    def read_df(bytes_object, **kwargs) -> pd.DataFrame: ...  # pragma: no cover

    @staticmethod
    @abstractmethod
    def write_df(df: pd.DataFrame, remote_path: str, **kwargs) -> None: ...  # pragma: no cover

    @staticmethod
    def verify_compression_and_extension(compression: str | None, filepath: str) -> None:
        """Check compression mode.

        Args:
            compression (str | None): Compression mode.
            filepath (str): File path.

        Raises:
            ValueError: If compression is not supported or if the file extension does not match the compression mode.
        """
        if not compression:
            return
        if compression not in AVAILABLE_COMPRESSIONS:
            raise ValueError(
                f"{compression} not supported. It must be in this list: {', '.join(AVAILABLE_COMPRESSIONS)}"
            )
        else:
            _, extension = path.splitext(filepath)
            available_extensions = AVAILABLE_COMPRESSIONS[compression]
            if available_extensions != ["*"] and extension not in available_extensions:
                raise ValueError(
                    f"A {compression}-compressed file must have an extension in this "
                    f"list: {', '.join(available_extensions)}"
                )


class CSVHandler(DataFrameHandler):
    @staticmethod
    def read_df(
        bytes_object: BufferType,
        compression: str = "infer",
        nrows: int | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Read a CSV file from a bytes object and return it as a DataFrame.

        Args:
            bytes_object (BufferType): The bytes object containing the CSV data.
            compression (str): The compression method to use (e.g., "gzip").
            nrows (int | None): The number of rows to read from the CSV file. If None, read all rows.
            **kwargs: Additional arguments to pass to the pandas read_csv function.

        Returns:
            pd.DataFrame: The DataFrame containing the data from the CSV file.
        """
        if nrows is not None and compression == "gzip":  # top N from gzip file
            return CSVHandler._read_df_nrows_gzip(bytes_object, nrows=nrows, **kwargs)
        elif nrows is not None:  # top N from file
            return CSVHandler._read_df_nrows(bytes_object, nrows=nrows, compression=compression, **kwargs)
        return CSVHandler._read_df(bytes_object, compression, **kwargs)  # default behaviour

    @staticmethod
    def write_df(df: pd.DataFrame, compression: str | None = None, **kwargs) -> io.BytesIO:
        """Write a DataFrame to a bytes object in CSV format.

        Args:
            df (pd.DataFrame): The DataFrame to write.
            compression (str | None): The compression method to use (e.g., "gzip").
            **kwargs: Additional arguments to pass to the pandas to_csv function.

        Returns:
            io.BytesIO: The bytes object containing the CSV data.
        """
        if not compression:
            buffer = io.StringIO()
            df.to_csv(buffer, **kwargs)
            buffer = io.BytesIO(buffer.getvalue().encode())
            return buffer

        elif compression in _SUPPORTED_COMPRESSIONS:
            buffer = io.BytesIO()
            df.to_csv(buffer, compression=compression, **kwargs)  # Only works for pandas >=1.2.0
            buffer.seek(0)
            return buffer
        else:
            raise ValueError(f"{compression} compression not supported, supported one are: {_SUPPORTED_COMPRESSIONS}")

    @staticmethod
    def _read_df(bytes_object: BufferType, compression: str | None = None, **kwargs) -> pd.DataFrame:
        buffer = io.BytesIO(bytes_object.read())
        return pd.read_csv(buffer, compression=compression, **kwargs)

    @staticmethod
    def _read_df_nrows(bytes_object: BufferType, nrows: int, compression: str | None = None, **kwargs) -> pd.DataFrame:
        content = []
        for idx, line in enumerate(bytes_object.iter_lines()):
            if idx > nrows:
                break
            content.append(line)
        return pd.read_csv(io.BytesIO(b"\n".join(content)), compression=compression, nrows=nrows, **kwargs)

    @staticmethod
    def _read_df_nrows_gzip(bytes_object: BufferType, nrows: int, **kwargs) -> pd.DataFrame:
        content = b""
        decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS | 16)
        header = kwargs.get("header", 0)
        n_lines = 0
        for chunk in bytes_object.iter_chunks():
            decompressed_chunk = decompressor.decompress(chunk)
            n_lines += decompressed_chunk.count(b"\n")
            content += decompressed_chunk
            if n_lines - (header == 0) > nrows:
                break
        return pd.read_csv(io.BytesIO(content), compression=None, nrows=nrows, **kwargs)


class ParquetHandler(DataFrameHandler):
    @staticmethod
    def read_df(bytes_object: BufferType, **kwargs) -> pd.DataFrame:
        """Read a Parquet file from a bytes object and return it as a DataFrame.

        Args:
            bytes_object (BufferType): The bytes object containing the Parquet data.
            **kwargs: Additional arguments to pass to the pandas read_parquet function.

        Returns:
            pd.DataFrame: The DataFrame containing the data from the Parquet file.
        """
        kwargs.pop("engine", None)
        reader = pa.BufferReader(bytes_object.read())
        return pd.read_parquet(reader, engine="pyarrow", **kwargs)

    @staticmethod
    def write_df(df: pd.DataFrame, compression: str | None = None, **kwargs) -> io.BytesIO:
        """Write a DataFrame to a bytes object in Parquet format.

        Args:
            df (pd.DataFrame): The DataFrame to write.
            compression (str | None): The compression method to use (e.g., "gzip").
            **kwargs: Additional arguments to pass to the pandas to_parquet function.

        Returns:
            io.BytesIO: The bytes object containing the Parquet data.
        """
        kwargs.pop("engine", None)
        buffer = io.BytesIO()
        df.to_parquet(buffer, engine="pyarrow", compression=compression, **kwargs)
        buffer.seek(0)
        return buffer


HANDLERS_MAP = {
    "csv": CSVHandler,
    "parquet": ParquetHandler,
}


def get_dataframe_handler(file_format: str = "csv") -> DataFrameHandler:
    """Get the appropriate DataFrame handler based on the file format.

    Args:
        file_format (str): The file format (e.g., "csv", "parquet").

    Returns:
        DataFrameHandler: The appropriate DataFrame handler class.

    Raises:
        ValueError: If the file format is not supported.
    """
    if file_format not in HANDLERS_MAP.keys():
        raise ValueError(f"{file_format} is not supported. Valid ones are {', '.join(HANDLERS_MAP)}.")

    return HANDLERS_MAP[file_format]



from botocore.paginate import PageIterator

SEPARATOR = "/"


def remove_separator_at_begin(path: str) -> str:
    """Remove the leading separator from the given path.

    Args:
        path (str): The given path.

    Returns:
        str: The path without the leading separator.
    """
    return path.lstrip(SEPARATOR) if path != SEPARATOR else path


def format_remote_path_folder_to_search(path: str) -> str:
    """Format the given path to be used in S3 search.

    Args:
        path (str): The given path.

    Returns:
        str: The formatted path.
    """
    path = remove_separator_at_begin(path)
    if not path.endswith(SEPARATOR) and not path.endswith("*"):
        path += SEPARATOR
    if "*" not in path:
        path += "*"
    return path


def get_folders_and_files(remote_dir_path: str, page_iterator: PageIterator) -> tuple[list[str], list[str]]:
    """Get the folders and files from the given page iterator.

    Args:
        remote_dir_path (str): The remote directory path.
        page_iterator (PageIterator): The page iterator.

    Returns:
        tuple[list[str], list[str]]: A tuple containing the list of folders and files.
    """
    wildcard = remote_dir_path[remote_dir_path.index("*") :]
    files: list[str] = []
    folders: list[str] = []
    for page in page_iterator:
        if "Contents" not in page and "CommonPrefixes" not in page:
            raise FileNotFoundError(f"The directory or object '{remote_dir_path}' doesn't exist.")

        folders.extend([element["Prefix"] for element in page.get("CommonPrefixes", [])])

        page_contents: list[dict] = page.get("Contents", [])

        all_filepaths: list[str] = []
        for element in page_contents:
            _current_element: str = element["Key"]

            if _current_element.endswith(SEPARATOR):
                folders.append(_current_element)
            else:
                all_filepaths.append(_current_element)

        files_filtered: list = _filter_files_according_to_wildcard(wildcard, all_filepaths)

        files.extend(files_filtered)

    folders.extend(_fill_folders_from_filepaths(files, remote_dir_path))

    return sorted(set(folders)), sorted(set(files))


def _filter_files_according_to_wildcard(wildcard: str, filepaths: list[str]) -> list[str]:
    """Find files with wildcard. Allowed wildcards are : path/to/*/[foo/bar]/[*|*.ext|file.ext]

    Args:
        wildcard (str): The wildcard to filter files.
        page_contents (List[Dict[str, str]]): Remote directory content.

    Returns:
        List[str]: Files to keep.
    """
    if not wildcard or wildcard == "*":
        return filepaths

    file_ext, directory = _split_and_format_wildcard(wildcard)
    return [file for file in filepaths if _file_must_be_included(file_ext, directory, file)]


def _file_must_be_included(file_extension: str, directory: str, file_path: str) -> bool:
    """Tell if the file must be included.

    Args:
        file_ext (str): Extension of file to keep.
        directory (str): Directory in wildcard if exists.
        file_path (str): File to filter path.

    Returns:
        bool: True if the file must be included, False otherwise.
    """
    _is_in_wildcard_with_required_extension: bool = (
        (f"{SEPARATOR}{directory}{SEPARATOR}" in file_path) or not directory
    ) and file_path.endswith(file_extension)

    return _is_in_wildcard_with_required_extension


def _split_and_format_wildcard(wildcard: str) -> tuple[str, str]:
    """Split the given wildcard and format it.

    Args:
        wildcard (str): The given wildcard.

    Returns:
        Tuple[str, str]: File extension and directory if exists.
    """
    formatted_wildcard: list = wildcard.replace("*", "").split(SEPARATOR)
    file_ext: str = formatted_wildcard[-1]
    directory: str = ""

    if len(formatted_wildcard) == 3:
        directory = formatted_wildcard[1]
    elif len(formatted_wildcard) == 2:
        directory = formatted_wildcard[0]

    return file_ext, directory


def _fill_folders_from_filepaths(files: list[str], remote_path: str) -> list[str]:
    """Used in case of unexpected behavior of boto3 paginator.
    This function fills folders list based on given filepaths information.

    Args:
        files (List[str]): Input file list.

    Returns:
        List[str]: Folder list.
    """
    if not remote_path.endswith(SEPARATOR):
        remote_path = f"{remote_path}{SEPARATOR}"

        if not remote_path.endswith("*"):
            remote_path = f"{remote_path}*"

    folders: list[str] = []
    prefix: str = remote_path[: remote_path.index("*")]

    for item in files:
        _splitted_path: list[str] = item.split(f"{prefix}{SEPARATOR}")[-1].split(SEPARATOR)
        parent_directory = _splitted_path[: len(_splitted_path) - 1]
        if not parent_directory:
            continue
        _all_subdir = f"{SEPARATOR}".join(parent_directory)
        _subdir_to_create: str = f"{_all_subdir}{SEPARATOR}"
        _equals_remote_path: bool = _subdir_to_create in prefix
        _is_already_registered: bool = _subdir_to_create in folders

        if _subdir_to_create and not _equals_remote_path and not _is_already_registered:
            folders.append(_subdir_to_create)

    return folders



import hashlib
import json
from configparser import ConfigParser, SectionProxy
from dataclasses import asdict, is_dataclass
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Any

import pandas as pd
import yaml

from .constants import EXT


class Extensions(str, Enum):
    YML = ".yml"
    YAML = ".yaml"
    CFG = ".cfg"
    INI = ".ini"
    JSON = ".json"


class TimeUnit(str, Enum):
    SECOND = "second"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"


def calculate_datetime_from_now(value: int, unit: TimeUnit) -> datetime:
    """
    Calculate datetime from now based on the given value and unit.

    Args:
        value (int): The value to calculate the datetime from now.
        unit (TimeUnit): The unit to calculate the datetime from now.
    Returns:
        datetime: The calculated datetime from now.
    """
    now = datetime.now()

    if unit == TimeUnit.SECOND:
        delta = timedelta(seconds=value)
    elif unit == TimeUnit.MINUTE:
        delta = timedelta(minutes=value)
    elif unit == TimeUnit.HOUR:
        delta = timedelta(hours=value)
    elif unit == TimeUnit.DAY:
        delta = timedelta(days=value)
    elif unit == TimeUnit.WEEK:
        delta = timedelta(weeks=value)
    else:  # pragma: no cover
        pass

    return now - delta


def get_datetime_from_filename(filename: Path | str) -> datetime:
    """
    Get datetime from the given filename.

    Args:
        filename (Path | str): The filename to get the datetime from.
    Returns:
        datetime: The datetime from the given filename.
    """
    filename_without_ext: str = str(filename).split(EXT)[0]
    extract_day, extract_time = filename_without_ext.split("_")[-1].split("T")
    extract_time = extract_time.replace("-", ":")
    return datetime.fromisoformat(f"{extract_day}T{extract_time}")


def make_json_serializable(obj: Any) -> Any:
    """
    Convert an object to a JSON-serializable format.
    This function handles various types of objects, including pandas DataFrames,
    dataclasses, dictionaries, lists, tuples, sets, and primitive types.

    Args:
        obj (Any): The object to convert to a JSON-serializable format.
    Returns:
        Any: The JSON-serializable representation of the object.
    """
    if isinstance(obj, pd.DataFrame):
        # Sort DataFrame by index and columns to ensure stable output
        # and reset index to avoid issues with non-serializable index types
        df = obj.sort_index(axis=1).sort_index().reset_index(drop=True)
        return df.to_dict(orient="records")  # Stable, JSON-compatible format
    elif is_dataclass(obj):
        return make_json_serializable(asdict(obj))
    elif isinstance(obj, dict):
        return {str(k): make_json_serializable(v) for k, v in sorted(obj.items())}
    elif isinstance(obj, list):
        return [make_json_serializable(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(make_json_serializable(v) for v in obj)
    elif isinstance(obj, set):
        return sorted(make_json_serializable(v) for v in obj)
    elif isinstance(obj, (int, float, str, bool)) or obj is None:
        return obj
    else:
        try:
            return make_json_serializable(vars(obj))
        except TypeError:
            return str(obj)


def hash_arguments(args: dict[str, Any]) -> str:
    """
    Hash the given arguments to create a unique identifier.
    This function converts the arguments to a JSON-serializable format,
    serializes them to a JSON string, and then computes the SHA-256 hash.

    Args:
        args (dict[str, Any]): The arguments to hash.
    Returns:
        str: The SHA-256 hash of the JSON-serializable arguments.
    """
    serializable_args = make_json_serializable(args)
    json_string = json.dumps(serializable_args, separators=(",", ":"), sort_keys=True)
    return hashlib.sha256(json_string.encode("utf-8")).hexdigest()


def load_file(file_path: str, key: str | None = None) -> dict | ConfigParser | SectionProxy:
    """
    Load a file into a dictionary or ConfigParser object.

    Args:
        file_path (str): The path to the file.
        key (str | None): The key to retrieve from the file. Defaults to None.

    Returns:
        dict | ConfigParser: The loaded file data.

    Raises:
        ValueError: If the file format is not supported.
        KeyError: If the key is not found in the file.
    """
    if file_path.endswith(Extensions.YML) or file_path.endswith(Extensions.YAML):
        file_dict = _load_yaml_file(file_path=file_path)
    elif file_path.endswith(Extensions.CFG) or file_path.endswith(Extensions.INI):
        file_dict = _load_cfg_file(file_path=file_path)
    elif file_path.endswith(Extensions.JSON):
        file_dict = _load_json_file(file_path=file_path)
    else:
        raise ValueError("File format not supported")

    try:
        return file_dict[key] if key else file_dict
    except KeyError:
        raise KeyError(f"Key {key} not found in {file_path}")


def _load_yaml_file(file_path: str) -> dict:
    """
    Load a yaml file into a dictionary

    Args:
        file_path (str): Path to the file

    Returns:
        dict: yaml file data
    """
    with open(file_path, "r") as stream:
        return yaml.safe_load(stream) or {}


def _load_cfg_file(file_path: str) -> ConfigParser:
    """
    Load a cfg file into a ConfigParser object

    Args:
        file_path (str): Path to the file

    Returns:
        ConfigParser: cfg file data
    """
    config_parser = ConfigParser(
        interpolation=None, converters={"list": json.loads, "dict": json.loads}
    )
    config_parser.read(file_path)
    return config_parser


def _load_json_file(file_path: str) -> dict:
    """
    Load a json file into a dictionary

    Args:
        file_path (str): Path to the file

    Returns:
        dict: json file data
    """
    with open(file_path, "r") as f:
        return json.load(f)



import os


def get_ca_bundle_path() -> str | bool:
    """Get the CA bundle path.

    Returns:
        str | bool: The path to the CA bundle certificates or False if not found.
    """
    if os.environ.get("REQUESTS_CA_BUNDLE"):
        return os.environ["REQUESTS_CA_BUNDLE"].strip()

    return False



from contextvars import ContextVar, Token
from typing import Any

_checkpoint_metadata: ContextVar[dict] = ContextVar("_checkpoint_metadata", default={})


def add_metadata(key: str, value: Any) -> None:
    """
    Add metadata to the checkpoint context.

    Args:
        key (str): The key for the metadata.
        value (Any): The value for the metadata.
    """
    metadata = _checkpoint_metadata.get()
    metadata[key] = value
    _checkpoint_metadata.set(metadata)


def add_multiple_metadata(metadata: dict[str, Any]) -> None:
    """
    Add multiple metadata to the checkpoint context.

    Args:
        metadata (dict): The metadata to add.
    """
    current_metadata = _checkpoint_metadata.get()
    current_metadata.update(metadata)
    _checkpoint_metadata.set(current_metadata)


def get_metadata() -> dict:
    """
    Get the metadata from the checkpoint context.

    Returns:
        dict: The metadata from the checkpoint context.
    """
    return _checkpoint_metadata.get()


def clear_metadata() -> Token:
    """
    Clear the metadata from the checkpoint context.
    """
    return _checkpoint_metadata.set({})


def restore_metadata(token: Token) -> None:
    """
    Restore the metadata in the checkpoint context.

    Args:
        token (Token): The token to restore the metadata.
    """
    _checkpoint_metadata.reset(token)



import boto3
from botocore.client import BaseClient, Config

from ..helpers.ssl import get_ca_bundle_path
from .config import S3Config


class get_s3_connection:
    """Open a new S3 connection.

    Use this function as a context manager:
        with get_s3_connection(...) as con:
            ...

    Args:
        config (S3Config): S3 configuration object.
    """

    def __init__(self, config: S3Config) -> None:
        self._config = config
        self._connection: BaseClient | None = None

    def open(self) -> None:
        """Open S3 connection."""
        self._connection = boto3.client(
            "s3",
            aws_access_key_id=self._config.access_key,
            aws_secret_access_key=self._config.secret_key,
            region_name=self._config.region_name,
            endpoint_url=self._config.endpoint_url,
            verify=get_ca_bundle_path() if not self._config.cert_file_path else self._config.cert_file_path,
            config=Config(
                signature_version=self._config.signature_version,
                s3={"addressing_style": "path" if self._config.force_path_style else "virtual"},
                request_checksum_calculation="when_required",
                response_checksum_validation="when_required",
            ),
        )

    def close(self) -> None:
        """Close ceph connection."""
        if self._connection:
            self._connection = None

    def is_open(self) -> bool:
        """Chek if connection is open."""
        return bool(self._connection)

    @property
    def connection(self) -> BaseClient:
        if self._connection:
            return self._connection

        return self._build_connection()

    @connection.setter
    def connection(self, _) -> None:
        raise ValueError("The connection attribute is immutable.")

    def _build_connection(self) -> BaseClient:
        if self._connection:
            return self._connection
        self.open()
        return self._connection

    def __enter__(self) -> BaseClient:
        self._build_connection()
        return self._connection

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()



from pydantic import BaseModel


class S3Config(BaseModel):
    access_key: str
    secret_key: str
    bucket_name: str
    region_name: str | None = None
    endpoint_url: str | None = None
    force_path_style: bool = True
    signature_version: str = "s3v4"
    cert_file_path: str | None = None



from configparser import SectionProxy
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any

import pandas as pd
import pytest

from resnap.helpers.utils import (
    TimeUnit,
    calculate_datetime_from_now,
    get_datetime_from_filename,
    hash_arguments,
    load_file,
)


@pytest.mark.parametrize(
    "value, unit, delta",
    [
        (10, TimeUnit.SECOND, timedelta(seconds=10)),
        (10, TimeUnit.MINUTE, timedelta(minutes=10)),
        (10, TimeUnit.HOUR, timedelta(hours=10)),
        (10, TimeUnit.DAY, timedelta(days=10)),
        (10, TimeUnit.WEEK, timedelta(weeks=10)),
    ],
)
def test_calculate_datetime_from_now(value: int, unit: TimeUnit, delta: timedelta) -> None:
    # Given
    now = datetime.now()

    # When
    result = calculate_datetime_from_now(value, unit)

    # Then
    expected = now - delta
    assert abs((result - expected).total_seconds()) < 1


@pytest.mark.parametrize(
    "filename, expected",
    [
        (
            "toto_2021-01-01T00-00-00.resnap",
            datetime.fromisoformat("2021-01-01T00:00:00"),
        ),
        (
            Path("toto_2021-01-01T00-00-00.resnap"),
            datetime.fromisoformat("2021-01-01T00:00:00"),
        ),
        (
            Path("toto/toto_2021-01-01T00-00-00.resnap"),
            datetime.fromisoformat("2021-01-01T00:00:00"),
        ),
    ],
)
def test_should_extract_datetime_from_filename(filename: Path | str, expected: datetime) -> None:
    # When
    result = get_datetime_from_filename(filename)

    # Then
    assert result == expected


@dataclass
class TestDataClass:
    arg_str: str
    arg_int: int
    arg_list: list[int]
    arg_df: pd.DataFrame | None = None


class TestClass:
    def __init__(self, arg_str: str, arg_int: int, arg_list: list[int], arg_df: pd.DataFrame | None = None) -> None:
        self.arg_str = arg_str
        self.arg_int = arg_int
        self.arg_list = arg_list
        self.arg_df = arg_df


@pytest.mark.parametrize(
    "arguments, expected",
    [
        (
            {"arg_str": "test", "arg_int": 42, "arg_list": [1, 2, 3]},
            "33e639b122556d8ad669ae7dcc9cf6576c0fdb6db6058acf0de5af199822b0be",
        ),
        (
            {"arg_int": 42, "arg_list": [1, 2, 3], "arg_str": "test"},
            "33e639b122556d8ad669ae7dcc9cf6576c0fdb6db6058acf0de5af199822b0be",
        ),
        (
            {"arg_int": 42, "arg_tuple": (1, 2, 3), "arg_str": "test"},
            "ac616acb116c3f94bce02223d7544c49c41cc431798fe3c39e38f9cefc03ac2b",
        ),
        (
            {"arg_int": 42, "arg_set": {1, 2, 3}, "arg_str": "test"},
            "f3868811a5ec3310128f870c3c25b05d8f32999696cfee883ed07b4862aefc63",
        ),
        (
            {"arg_df": pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]}), "arg_str": "test"},
            "309f001ffe0ee049f7fa34f32f7668424e589a73795257a20e652dfaf78b73e8",
        ),
        (
            {"arg_df": pd.DataFrame({"A": [3, 2, 1], "B": [6, 5, 4]}), "arg_str": "test"},
            "cbaa16e213cf07f92cdd588522e8ffc458bc14b8acf89fce7196539ef1116e50",
        ),
        (
            {"args_dataclass": TestDataClass("test", 42, [1, 2, 3], arg_df=pd.DataFrame({"A": [1, 2], "B": [3, 4]}))},
            "cf807ac07815f225022df95c32adec700dc71e7a11331187aa1f0c74f8aa8237",
        ),
        (
            {"args_class": TestClass("test", 42, [1, 2, 3], arg_df=pd.DataFrame({"A": [1, 2], "B": [3, 4]}))},
            "54af02a514923a8f229e1bf8c39fd5d63e7b2d939a76ea0bf585be82284ffe13",
        ),
        (
            {"arg_slice": slice(1, 5, 2), "arg_str": "test"},
            "1bdd21fe2e391a88faf9671df103d38d55bcc6f6e17ef3b6d0e373ba89e8bf5c",
        ),
    ],
)
def test_should_hash_arguments(arguments: dict[str, Any], expected: str) -> None:
    # When
    result = hash_arguments(arguments)

    # Then
    assert result == expected


@pytest.mark.parametrize(
    "file_path, result_type",
    [
        ("test-secrets.json", dict),
        ("test-secrets.yml", dict),
        ("test-secrets.cfg", SectionProxy),
    ],
)
def test_should_load_yaml_file(file_path: str, result_type: dict | SectionProxy) -> None:
    # Given
    file_path = Path(__file__).parent.parent / "data" / "config" / file_path

    # When
    result = load_file(str(file_path), key="s3")

    # Then
    assert isinstance(result, result_type)
    assert result["server"] == "server"
    assert result["access_key"] == "access_key"
    assert result["secret_key"] == "secret_key"
    assert result["bucket"] == "bucket"


def test_should_raise_value_error_when_file_format_is_not_supported() -> None:
    # Given
    file_path = Path(__file__).parent.parent / "data" / "config" / "test-secrets.txt"

    # When / Then
    with pytest.raises(ValueError, match="File format not supported"):
        load_file(str(file_path))


def test_should_raise_key_error_when_key_is_not_found() -> None:
    # Given
    file_path = Path(__file__).parent.parent / "data" / "config" / "test-secrets.cfg"

    # When
    with pytest.raises(KeyError) as e:
        load_file(str(file_path), key="non_existent")

    # Then
    assert "Key non_existent not found" in str(e.value)



"""Defines the public resnap interface"""

from .decorators import async_resnap, resnap
from .exceptions import ResnapError
from .factory import set_resnap_service
from .helpers.context import add_metadata, add_multiple_metadata
from .services.service import ResnapService
from .version import VERSION

__version__ = VERSION
__all__ = (
    # decorators
    "resnap",
    "async_resnap",
    # exceptions
    "ResnapError",
    # factory
    "set_resnap_service",
    # metadata
    "add_metadata",
    "add_multiple_metadata",
    # services
    "ResnapService",
    # version
    "__version__",
    "VERSION",
)



import pandas as pd

FILES = [
    "Pipeline._first_method",
    "Pipeline._second_method",
    "Pipeline._third_method",
    "BasicClass.get_param_value",
    "BasicClass.generate_dataframe",
    "a_function",
]
TEST_CONFIG_GLOBAL = {
    "first_method_argument": pd.DataFrame(
        {
            "A": [1, 2, 3],
            "B": [4, 5, 6],
            "C": [7, 8, 9],
        }
    ),
    "second_method_argument": "titi",
    "third_method_argument": "titi",
    "BasicClass_get_param_value_argument": "titi",
    "BasicClass_generate_dataframe_argument": "titi",
    "a_function_argument": True,
}



from datetime import datetime
from typing import Any
from unittest.mock import MagicMock, call

import freezegun
import pandas as pd
import pytest

from resnap.helpers.metadata import MetadataFail, MetadataSuccess
from resnap.helpers.singleton import SingletonABCMeta
from resnap.helpers.status import Status
from resnap.helpers.utils import hash_arguments
from resnap.services.local_service import LocalResnapService
from tests.builders.config_builder import ConfigBuilder


@pytest.fixture(autouse=True)
def reset_singletons() -> None:
    SingletonABCMeta._instances = {}


@pytest.fixture(autouse=True)
def mock_read_parquet_to_dataframe(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._read_parquet_to_dataframe"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_read_csv_to_dataframe(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._read_csv_to_dataframe"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_read_pickle(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._read_pickle"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_read_text(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._read_text"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_read_json(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._read_json"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_save_dataframe_to_parquet(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._save_dataframe_to_parquet"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_save_dataframe_to_csv(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._save_dataframe_to_csv"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_save_to_pickle(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._save_to_pickle"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_save_to_text(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._save_to_text"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_save_to_json(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._save_to_json"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_write_metadata(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._write_metadata"
    )
    return mock


@pytest.fixture(autouse=True)
def mock_create_folder(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.local_service.LocalResnapService._create_folder"
    )
    return mock


class TestServicce:
    hashed_arguments: str = hash_arguments({"toto": "toto"})

    @pytest.mark.parametrize("enabled", [True, False])
    def test_should_return_if_service_is_enabled(self, enabled: bool) -> None:
        # Given
        service = LocalResnapService(
            config=ConfigBuilder.a_config().is_enabled(enabled).build()
        )
        assert service.is_enabled == enabled

    @pytest.mark.parametrize(
        "output_path, output_folder, expected",
        [
            ("", "", "test_2021-01-01T00-00-00.resnap"),
            ("output", "", "output/test_2021-01-01T00-00-00.resnap"),
            ("output", "test", "output/test/test_2021-01-01T00-00-00.resnap"),
        ],
    )
    def test_should_return_metadata_path(
        self, output_path: str, output_folder: str, expected: str
    ) -> None:
        # Given
        service = LocalResnapService(
            config=ConfigBuilder.a_config().with_output_base_path(output_path).build()
        )
        func_name = "test"
        event_time = datetime.fromisoformat("2021-01-01T00:00:00")

        # When
        result = service.metadata_path(func_name, event_time, output_folder)

        # Then
        assert result == expected

    @pytest.mark.parametrize(
        "output_path, output_folder, expected",
        [
            ("", "", "test_2021-01-01T00-00-00.resnap.ext"),
            ("output", "", "output/test_2021-01-01T00-00-00.resnap.ext"),
            ("output", "test", "output/test/test_2021-01-01T00-00-00.resnap.ext"),
        ],
    )
    def test_should_return_result_path(
        self, output_path: str, output_folder: str, expected: str
    ) -> None:
        # Given
        service = LocalResnapService(
            config=ConfigBuilder.a_config().with_output_base_path(output_path).build()
        )
        func_name = "test"
        event_time = datetime.fromisoformat("2021-01-01T00:00:00")

        # When
        result = service.result_path(func_name, event_time, output_folder, "ext")

        # Then
        assert result == expected

    @pytest.mark.parametrize(
        "result_path, expected_mock, result_type",
        [
            ("test_2021-01-01T00-00-00.resnap.pkl", "mock_read_pickle", "str"),
            (
                "test_2021-01-01T00-00-00.resnap.parquet.gz",
                "mock_read_parquet_to_dataframe",
                "pd.DataFrame",
            ),
            (
                "test_2021-01-01T00-00-00.resnap.csv",
                "mock_read_csv_to_dataframe",
                "pd.DataFrame",
            ),
            ("test_2021-01-01T00-00-00.resnap.txt", "mock_read_text", "str"),
            ("test_2021-01-01T00-00-00.resnap.json", "mock_read_json", "str"),
        ],
    )
    def test_should_read_result(
        self,
        result_path: str,
        expected_mock: str,
        result_type: str,
        request: type[pytest.FixtureRequest],
    ) -> None:
        # Given
        mock: MagicMock = request.getfixturevalue(expected_mock)
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hash_arguments({}),
            result_path=result_path,
            result_type=result_type,
        )

        # When
        service.read_result(metadata)

        # Then
        mock.assert_called_once_with(metadata.result_path)

    @pytest.mark.parametrize(
        "result_type, result",
        [
            ("str", "test"),
            ("int", 10),
            ("float", 5.2),
            ("bool", True),
        ],
    )
    def test_should_cast_result_if_file_is_text(
        self, result_type: str, result: Any, mock_read_text: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hash_arguments({}),
            result_path="test_2021-01-01T00-00-00.resnap.txt",
            result_type=result_type,
        )
        mock_read_text.return_value = str(result)

        # When
        res = service.read_result(metadata)

        # Then
        assert res == result

    @pytest.mark.parametrize(
        "result_path, result_type",
        [
            ("test_2021-01-01T00-00-00.resnap.toto", "str"),
            ("test_2021-01-01T00-00-00.resnap.toto", "pd.DataFrame"),
        ],
    )
    def test_should_raise_not_implemented_error_when_reading_unknown_result_type(
        self,
        result_path: str,
        result_type: str,
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            hashed_arguments=hash_arguments({}),
            result_path=result_path,
            result_type=result_type,
        )

        # When / Then
        with pytest.raises(
            NotImplementedError, match=f"Unsupported result type: {result_type}"
        ):
            service.read_result(metadata)

    @freezegun.freeze_time("2021-01-01")
    @pytest.mark.parametrize(
        "func_result, output_format, output_folder, expected_mock, expected_result_ext",
        [
            ("toto", None, "", "mock_save_to_pickle", "pkl"),
            ("toto", "pkl", "", "mock_save_to_pickle", "pkl"),
            (10, "txt", "", "mock_save_to_text", "txt"),
            (False, None, "", "mock_save_to_pickle", "pkl"),
            ({"test": 1}, "json", "", "mock_save_to_json", "json"),
            ({"test": 1}, "json", "toto", "mock_save_to_json", "json"),
            (
                pd.DataFrame(
                    {
                        "A": [1, 2, 3],
                        "B": [4, 5, 6],
                        "C": [7, 8, 9],
                    }
                ),
                None,
                "",
                "mock_save_dataframe_to_parquet",
                "parquet.gz",
            ),
            (
                pd.DataFrame(
                    {
                        "A": [1, 2, 3],
                        "B": [4, 5, 6],
                        "C": [7, 8, 9],
                    }
                ),
                "csv",
                "",
                "mock_save_dataframe_to_csv",
                "csv",
            ),
        ],
    )
    def test_should_save_result(
        self,
        func_result: Any,
        output_format: str | None,
        output_folder: str,
        expected_mock: str,
        expected_result_ext: str,
        request: type[pytest.FixtureRequest],
    ) -> None:
        # Given
        mock: MagicMock = request.getfixturevalue(expected_mock)
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        expected_event_time: datetime = datetime.fromisoformat("2021-01-01T00:00:00")
        output_path = f"{output_folder}/" if output_folder else ""
        expected_result_path: str = (
            f"{output_path}test_2021-01-01T00-00-00.resnap.{expected_result_ext}"
        )

        # When
        result_path, event_time = service.save_result(
            "test", func_result, output_folder, output_format
        )

        # Then
        mock.assert_called_once_with(
            result=func_result, result_path=expected_result_path
        )
        assert result_path == expected_result_path
        assert event_time == expected_event_time

    def test_should_save_success_metadata(self, mock_write_metadata: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        expected_event_time: datetime = datetime.fromisoformat("2021-01-01T00:00:00")
        expected_result_path: str = "test_2021-01-01T00-00-00.resnap.pkl"
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=expected_event_time,
            hashed_arguments=self.hashed_arguments,
            result_path=expected_result_path,
            result_type="str",
            extra_metadata={},
        )

        # When
        service.save_success_metadata(
            "test",
            "",
            self.hashed_arguments,
            expected_event_time,
            expected_result_path,
            "str",
            extra_metadata={},
        )

        # Then
        mock_write_metadata.assert_called_once_with(
            "test_2021-01-01T00-00-00.resnap", metadata
        )

    def test_should_save_failed_metadata(self, mock_write_metadata: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        expected_event_time: datetime = datetime.fromisoformat("2021-01-01T00:00:00")
        metadata = MetadataFail(
            status=Status.FAIL,
            event_time=expected_event_time,
            hashed_arguments=self.hashed_arguments,
            error_message="error",
            data={"id": 1},
            extra_metadata={},
        )

        # When
        service.save_failed_metadata(
            "test",
            "",
            self.hashed_arguments,
            expected_event_time,
            error_message="error",
            data={"id": 1},
            extra_metadata={},
        )

        # Then
        mock_write_metadata.assert_called_once_with(
            "test_2021-01-01T00-00-00.resnap",
            metadata,
        )

    @pytest.mark.parametrize(
        "path, output_folder",
        [
            ("", ""),
            ("results", ""),
            ("results", "test"),
        ],
    )
    def test_should_create_folder(
        self, path: str, output_folder: str, mock_create_folder: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(
            config=ConfigBuilder.a_config().with_output_base_path(path).build()
        )
        expected_calls = []
        if path:
            expected_calls.append(call("", path))
        if output_folder:
            expected_calls.append(call(path, output_folder))

        # When
        service.create_output_folder(output_folder)

        # Then
        mock_create_folder.assert_has_calls(calls=expected_calls, any_order=True)



import io

import pandas as pd

from .config import S3Config
from .connection import get_s3_connection
from .dataframes import get_dataframe_handler
from .tools import SEPARATOR, format_remote_path_folder_to_search, get_folders_and_files, remove_separator_at_begin

_UNDEFINED_VALUE = object()


class S3Client:
    """S3Client is a class that provides methods to interact with Amazon S3 or Ceph.
    It allows uploading, downloading, listing, and deleting files and directories in S3.

    Attributes:
        config (S3Config): Configuration object containing S3 connection details.
    """

    def __init__(self, config: S3Config) -> None:
        self.config = config
        self.bucket_name = config.bucket_name

    def upload_file(self, local_path_or_fileobj: str | io.FileIO | io.BytesIO, remote_path: str) -> None:
        """Upload a file to S3.

        Args:
            local_path_or_fileobj (str | io.FileIO | io.BytesIO): Local file path or file-like object to upload.
            remote_path (str): S3 path (key) where the file will be uploaded.
        """
        remote_path = remove_separator_at_begin(remote_path)
        with get_s3_connection(self.config) as connection:
            if isinstance(local_path_or_fileobj, str):
                with open(local_path_or_fileobj, "rb") as file:
                    connection.upload_fileobj(file, self.bucket_name, remote_path)
            else:
                if isinstance(local_path_or_fileobj, io.BytesIO):
                    local_path_or_fileobj.seek(0)  # re-init the buffer read position
                connection.upload_fileobj(local_path_or_fileobj, self.bucket_name, remote_path)

    def download_file(self, local_path_or_fileobj: str | io.FileIO | io.BytesIO, remote_path: str) -> None:
        """Download a file from S3.

        Args:
            local_path_or_fileobj (str | io.FileIO | io.BytesIO): Local file path or file-like object to download to.
            remote_path (str): S3 path (key) of the file to download.
        """
        remote_path = remove_separator_at_begin(remote_path)
        with get_s3_connection(self.config) as connection:
            if isinstance(local_path_or_fileobj, str):
                with open(local_path_or_fileobj, "wb") as file:
                    connection.download_fileobj(self.bucket_name, remote_path, file)
            else:
                if isinstance(local_path_or_fileobj, io.BytesIO):
                    local_path_or_fileobj.seek(0)  # re-init the buffer read position
                connection.download_fileobj(self.bucket_name, remote_path, local_path_or_fileobj)

    def list_objects(self, remote_dir_path: str = "", recursive: bool = False) -> list[str]:
        """List objects in a directory in S3.
        Args:
            remote_dir_path (str): The S3 path to the directory.
            recursive (bool): Whether to list objects recursively.

        Returns:
            List[str]: A list of object keys in the specified directory.
        """
        folders, files = self.list_folders_and_files(remote_dir_path, recursive)
        return sorted(folders + files)

    def list_folders_and_files(self, remote_dir_path: str = "", recursive: bool = False) -> tuple[list[str], list[str]]:
        """List folders and files in a directory in S3.

        Args:
            remote_dir_path (str): The S3 path to the directory.
            recursive (bool): Whether to list objects recursively.

        Returns:
            tuple[list[str], list[str]]: A tuple containing two lists: folders and files.
        """
        remote_dir_path = format_remote_path_folder_to_search(remote_dir_path)
        prefix = remote_dir_path[: remote_dir_path.index("*")]
        operation_parameters = {
            "Bucket": self.bucket_name,
            "Prefix": prefix,
        }
        if not recursive:
            operation_parameters["Delimiter"] = SEPARATOR

        with get_s3_connection(self.config) as connection:
            paginator = connection.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(**operation_parameters)
            folders, files = get_folders_and_files(remote_dir_path, page_iterator)
            return folders, files

    def delete_object(self, key: str) -> None:
        """Delete an object from S3.

        Args:
            key (str): The S3 key of the object to delete.
        """
        with get_s3_connection(self.config) as connection:
            connection.delete_object(Bucket=self.bucket_name, Key=key)

    def delete_objects(self, keys: list[str]) -> None:
        """Delete multiple objects from S3.

        Args:
            keys (list[str]): A list of S3 keys of the objects to delete.
        """
        objects = [{"Key": key} for key in keys]
        if not objects:
            return
        with get_s3_connection(self.config) as connection:
            connection.delete_objects(Bucket=self.bucket_name, Delete={"Objects": objects})

    def object_exists(self, remote_path: str) -> bool:
        """Check if an object exists in S3.

        Args:
            remote_path (str): The S3 path (key) of the object to check.

        Returns:
            bool: True if the object exists, False otherwise.
        """
        remote_path = remove_separator_at_begin(remote_path)
        with get_s3_connection(self.config) as connection:
            try:
                if remote_path == SEPARATOR:
                    connection.get_bucket_acl(Bucket=self.bucket_name)
                elif remote_path.endswith(SEPARATOR):
                    connection.get_object(Bucket=self.bucket_name, Key=remote_path)
                else:
                    connection.head_object(Bucket=self.bucket_name, Key=remote_path)
                return True
            except Exception:
                return False

    def mkdir(self, path: str) -> None:
        """Create a directory in S3. In S3, directories are represented by keys that end with a '/'.

        Args:
            path (str): The path to create.
        """
        if not path.endswith(SEPARATOR):
            path += SEPARATOR
        self.upload_file(io.BytesIO(), path)

    def rmdir(self, path: str) -> None:
        """Remove a directory in S3. In S3, directories are represented by keys that end with a '/'.

        Args:
            path (str): The path to remove.
        """
        path = format_remote_path_folder_to_search(path)
        folders, files = self.list_folders_and_files(path, recursive=True)
        self.delete_objects(files)
        self.delete_objects(folders)

    def get_df_from_file(self, remote_path: str, file_format: str = "csv", **kwargs) -> pd.DataFrame | None:
        """Read a file from S3 and return it as a DataFrame.

        Args:
            remote_path (str): The S3 path to the file.
            file_format (str): The format of the file (e.g., "csv", "parquet").

        Returns:
            pd.DataFrame: The DataFrame containing the data from the file.
        """
        df_handler = get_dataframe_handler(file_format)
        df_handler.verify_compression_and_extension(kwargs.get("compression"), remote_path)

        remote_path = remove_separator_at_begin(remote_path)
        if not self.object_exists(remote_path):
            raise FileNotFoundError(f"File {remote_path} not exists in bucket {self.bucket_name}")

        with get_s3_connection(self.config) as connection:
            bytes_object = connection.get_object(Bucket=self.bucket_name, Key=remote_path)["Body"]
            return df_handler.read_df(bytes_object, **kwargs)

    def push_df_to_file(
        self,
        df: pd.DataFrame,
        remote_path: str,
        compression: str | None = _UNDEFINED_VALUE,
        file_format: str = "csv",
        **kwargs,
    ) -> None:
        """Save a DataFrame to a file in S3.

        Args:
            df (pd.DataFrame): The DataFrame to save.
            remote_path (str): The S3 path where the file will be saved.
            compression (str | None): The compression method to use (e.g., "gzip").
            file_format (str): The format of the file (e.g., "csv", "parquet").
        """
        if compression is _UNDEFINED_VALUE:
            compression = "snappy" if file_format == "parquet" else None

        df_handler = get_dataframe_handler(file_format)
        df_handler.verify_compression_and_extension(compression, remote_path)

        remote_path = remove_separator_at_begin(remote_path)
        bytes_object = df_handler.write_df(df, compression=compression, **kwargs)
        self.upload_file(bytes_object, remote_path)



import logging
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from resnap.helpers.config import Config
from resnap.settings import get_config_data, get_config_file_path


@pytest.fixture
def mock_os_listdir(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("os.listdir")
    return mock


@pytest.fixture(autouse=True)
def mock_os_getcwd(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("os.getcwd", return_value="tests")
    return mock


@pytest.fixture
def mock_get_config_file_path(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.settings.get_config_file_path")
    return mock


@pytest.fixture(autouse=True)
def mock_toml_load(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("toml.load")
    return mock


@pytest.mark.parametrize(
    "file_path, config_data, expected",
    [
        pytest.param(
            "tests/pyproject.toml",
            {"tool": {"resnap": {"enabled": True, "output_base_path": "resnap"}}},
            Config(enabled=True, output_base_path="resnap"),
            id="find config in pyproject.toml",
        ),
        pytest.param(
            "tests/pyproject.toml",
            {"toto": {"test": True}},
            Config(enabled=True),
            id="don't find tool config in pyproject.toml",
        ),
        pytest.param(
            "tests/pyproject.toml",
            {"tool": {"test": True}},
            Config(enabled=True),
            id="don't find resnap config in pyproject.toml",
        ),
    ],
)
def test_should_get_config_data(
    file_path: str,
    config_data: dict,
    expected: Config,
    mock_get_config_file_path: MagicMock,
    mock_toml_load: MagicMock,
) -> None:
    # Given
    mock_get_config_file_path.return_value = file_path
    mock_toml_load.return_value = config_data

    # When
    config = get_config_data()

    # Then
    assert isinstance(config, Config)
    assert config == expected


def test_should_not_call_get_config_file_path_if_file_path_is_provided(
    mock_get_config_file_path: MagicMock, mock_toml_load: MagicMock,
) -> None:
    # Given
    file_path = "tests/pyproject.toml"
    mock_toml_load.return_value = {"tool": {"resnap": {"enabled": True, "output_base_path": "resnap"}}}
    expected = Config(enabled=True, output_base_path="resnap")

    # When
    config = get_config_data(file_path)

    # Then
    assert isinstance(config, Config)
    assert config == expected
    mock_get_config_file_path.assert_not_called()


@pytest.mark.parametrize("file_path", ["", "test.txt"])
def test_should_return_default_config_data_with_warning(file_path: str, mock_get_config_file_path: MagicMock) -> None:
    # Given
    mock_get_config_file_path.return_value = file_path

    # When
    with patch.object(logging.getLogger("resnap"), "warning") as mock_warning:
        config = get_config_data()
        mock_warning.assert_called_once_with(f"Unsupported file type: {file_path}.")
        assert config == Config(enabled=True)


@pytest.mark.parametrize(
    "listdir_result, expected",
    [
        (["pyproject.toml"], "tests/pyproject.toml"),
        (["other_file.txt"], ""),
        (["another_file.txt", "pyproject.toml"], "tests/pyproject.toml"),
        ([], ""),
    ],
)
def test_should_get_config_file_path(listdir_result: list[str], expected: str, mock_os_listdir: MagicMock) -> None:
    # Given
    mock_os_listdir.return_value = listdir_result

    # When
    config_file_path = get_config_file_path()

    # Then
    assert Path(config_file_path) == Path(expected)



import sys

print(f"Running smoke test with Python {sys.version}")

try:
    import resnap
except ImportError as e:
    raise RuntimeError("❌ Échec de l'import du module principal.") from e

# Vérifions qu’une fonction ou classe clé est présente
assert hasattr(resnap, "resnap"), "❌ La fonction 'resnap' est manquante dans le package."

print("✅ Smoke test passed.")



from dependency_injector.containers import DeclarativeContainer
from dependency_injector.providers import Configuration, Singleton
from pipeline import Pipeline


class JobContainer(DeclarativeContainer):
    job_config: Configuration = Configuration()
    test = job_config.test
    pipeline: Pipeline = Singleton(Pipeline)



import asyncio
import threading
from typing import Any

import pytest

from resnap.helpers.context import (
    add_metadata,
    add_multiple_metadata,
    clear_metadata,
    get_metadata,
    restore_metadata,
)


def test_add_and_get_metadata() -> None:
    # Given
    clear_metadata()

    # When / Then
    add_metadata("key1", "value1")
    metadata = get_metadata()

    assert isinstance(metadata, dict)
    assert metadata == {"key1": "value1"}

    add_metadata("key2", 42)
    metadata = get_metadata()
    assert metadata == {"key1": "value1", "key2": 42}
    assert len(metadata) == 2

    # Finally
    clear_metadata()


def test_add_and_get_multiple_metadata() -> None:
    # Given
    clear_metadata()

    # When / Then
    add_metadata("key1", "value0")
    metadata = {
        "key1": "value1",
        "key2": 42,
        "key3": 3.14,
    }
    add_multiple_metadata(metadata)
    result = get_metadata()

    assert isinstance(result, dict)
    assert result == metadata

    # Finally
    clear_metadata()


def test_clear_metadata() -> None:
    # Given
    add_metadata("key", "value")
    metadata = get_metadata()
    assert metadata == {"key": "value"}

    # When / Then
    clear_metadata()
    metadata = get_metadata()
    assert metadata == {}


def test_clear_and_restore_metadata() -> None:
    add_metadata("a", 1)
    assert get_metadata() == {"a": 1}

    token = clear_metadata()
    assert get_metadata() == {}

    add_metadata("b", 2)
    assert get_metadata() == {"b": 2}

    restore_metadata(token)
    assert get_metadata() == {"a": 1}


async def set_and_get_metadata_async(key, value) -> dict[str, Any]:
    clear_metadata()
    add_metadata(key, value)
    await asyncio.sleep(0.1)
    return get_metadata()


@pytest.mark.asyncio
async def test_async_metadata_isolation() -> None:
    result1, result2 = await asyncio.gather(
        set_and_get_metadata_async("a", 1), set_and_get_metadata_async("b", 2)
    )

    assert result1 == {"a": 1}
    assert result2 == {"b": 2}


def set_and_get_metadata_thread(key, value, result_holder, index) -> None:
    clear_metadata()
    add_metadata(key, value)
    result_holder[index] = get_metadata()


def test_thread_metadata_isolation() -> None:
    result_holder = [None, None]

    t1 = threading.Thread(
        target=set_and_get_metadata_thread, args=("x", "foo", result_holder, 0)
    )
    t2 = threading.Thread(
        target=set_and_get_metadata_thread, args=("y", "bar", result_holder, 1)
    )

    t1.start()
    t2.start()
    t1.join()
    t2.join()

    assert result_holder[0] == {"x": "foo"}
    assert result_holder[1] == {"y": "bar"}



import pytest

from resnap.boto.tools import format_remote_path_folder_to_search, get_folders_and_files, remove_separator_at_begin


@pytest.mark.parametrize(
    "path, expected",
    [
        pytest.param("/test/path", "test/path", id="leading_slash"),
        pytest.param("test/path/", "test/path/", id="no_leading_slash"),
        pytest.param("/", "/", id="only_slash"),
    ],
)
def test_remove_separator_at_begin(path: str, expected: str) -> None:
    # When
    result = remove_separator_at_begin(path)

    # Then
    assert result == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        pytest.param("/test/path", "test/path/*", id="no_trailing_slash"),
        pytest.param("test/path/", "test/path/*", id="trailing_slash"),
        pytest.param("/test/path/*", "test/path/*", id="wildcard_in_path"),
    ],
)
def test_format_remote_path_folder_to_search(path: str, expected: str) -> None:
    # When
    result = format_remote_path_folder_to_search(path)

    # Then
    assert result == expected


@pytest.mark.parametrize(
    "path, iterator, expected_result",
    [
        pytest.param(
            "test/path/*",
            [
                {
                    "Contents": [
                        {"Key": "test/path/file1.txt"},
                        {"Key": "test/path/sub_path/"},
                    ]
                },
                {"CommonPrefixes": [{"Prefix": "test/path/folder1/"}]},
                {
                    "Contents": [
                        {"Key": "test/path/file2.txt"},
                        {"Key": "test/path/folder2/file2.txt"},
                    ]
                },
            ],
            (
                [
                    "test/path/folder1/",
                    "test/path/folder2/",
                    "test/path/sub_path/",
                ],
                [
                    "test/path/file1.txt",
                    "test/path/file2.txt",
                    "test/path/folder2/file2.txt",
                ],
            ),
            id="folders_and_files_without_wildcard",
        ),
        pytest.param(
            "test/path/*.txt",
            [
                {
                    "Contents": [
                        {"Key": "test/path/file1.txt"},
                        {"Key": "test/path/sub_path/"},
                    ]
                },
                {"CommonPrefixes": [{"Prefix": "test/path/folder1/"}]},
                {
                    "Contents": [
                        {"Key": "test/path/file2.csv"},
                        {"Key": "test/path/folder2/file2.json"},
                    ]
                },
            ],
            (
                [
                    "test/path/folder1/",
                    "test/path/sub_path/",
                ],
                [

                    "test/path/file1.txt",
                ],
            ),
            id="folders_and_files_with_wildcard_of_selected_type",
        ),
        pytest.param(
            "test/path/*/toto/*.txt",
            [
                {"Contents": [{"Key": "test/path/file1.txt"}]},
                {
                    "Contents": [
                        {"Key": "test/path/file2.csv"},
                        {"Key": "test/path/folder2/file2.json"},
                    ]
                },
                {
                    "Contents": [
                        {"Key": "test/path/toto/file2.csv"},
                        {"Key": "test/path/toto/file2.txt"},
                    ]
                },
            ],
            (
                [
                    "test/path/toto/",
                ],
                [
                    "test/path/toto/file2.txt",
                ],
            ),
            id="folders_and_files_with_wildcard_of_selected_type_in_one_directory",
        ),
        pytest.param(
            "test/path/*toto/*.txt",
            [
                {"Contents": [{"Key": "test/path/file1.txt"}]},
                {
                    "Contents": [
                        {"Key": "test/path/file2.csv"},
                        {"Key": "test/path/folder2/file2.json"},
                    ]
                },
                {
                    "Contents": [
                        {"Key": "test/path/toto/file2.csv"},
                        {"Key": "test/path/toto/file2.txt"},
                    ]
                },
            ],
            (
                [
                    "test/path/toto/",
                ],
                [
                    "test/path/toto/file2.txt",
                ],
            ),
            id="folders_and_files_with_wildcard_of_selected_type_in_one_directory_missing_slash",
        ),
        pytest.param(
            "/*",
            [
                {"Contents": [{"Key": "file1.txt"}]},
                {"CommonPrefixes": [{"Prefix": "folder1/"}]},
                {"Contents": [{"Key": "folder1/file2.csv"}]},
            ],
            (
                [
                    "folder1/",
                ],
                [
                    "file1.txt",
                    "folder1/file2.csv",
                ],
            ),
            id="folders_and_files_with_wildcard_in_root",
        ),
        pytest.param(
            "test/path/*",
            [
                {"Contents": []},
                {"CommonPrefixes": []},
            ],
            ([], []),
            id="empty_iterator",
        ),
    ],
)
def test_should_get_folders_and_files(path: str, iterator: list, expected_result: list[str]) -> None:
    # When
    result = get_folders_and_files(path, iterator)

    # Then
    assert result == expected_result


def test_should_raise_if_not_contents() -> None:
    # Given
    path = "test/path/*"
    iterator = [{"test": False}]

    # When / Then
    with pytest.raises(FileNotFoundError, match=r"The directory or object 'test/path/\*' doesn't exist."):
        get_folders_and_files(path, iterator)



from enum import Enum

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

from .utils import TimeUnit


class Services(str, Enum):
    LOCAL = "local"
    S3 = "s3"


class Config(BaseModel):
    enabled: bool
    save_to: Services = Services.LOCAL
    output_base_path: str = ""
    secrets_file_name: str = ""
    enable_remove_old_files: bool = False
    max_history_files_length: int = Field(gt=0, default=3)
    max_history_files_time_unit: TimeUnit = TimeUnit.DAY

    @model_validator(mode="after")
    def check_secrets_file_name(self) -> Self:
        if self.enabled and self.save_to != Services.LOCAL and not self.secrets_file_name:
            raise ValueError(f"secrets_file_name is required when save_to is {self.save_to.value}")
        return self



import re
import sys
from unittest.mock import MagicMock, patch

import pytest

from resnap import factory
from resnap.services.boto_service import BotoResnapService
from resnap.services.local_service import LocalResnapService
from resnap.services.service import ResnapService
from resnap.settings import get_config_data


@pytest.fixture(autouse=True)
def reset_factory_globals() -> None:
    factory._resnap_config = get_config_data()
    factory._service = None


def test_should_set_service() -> None:
    # Given
    custom_service = MagicMock(spec=ResnapService, name="CustomResnapService")

    # When
    factory.set_resnap_service(custom_service)

    # Then
    assert factory._service == custom_service
    assert isinstance(factory._service, ResnapService)


def test_should_not_set_service_if_not_resnap_service() -> None:
    # Given
    custom_service = MagicMock(name="CustomResnapService")

    # When / Then
    with pytest.raises(
        TypeError,
        match=re.escape(f"Expected ResnapService, got {type(custom_service)}"),
    ):
        factory.set_resnap_service(custom_service)


def test_should_raise_if_service_if_not_implemeted() -> None:
    # Given
    factory._resnap_config.save_to = "not_implemented"

    # When / Then
    with pytest.raises(
        NotImplementedError,
        match=re.escape("Resnap service not_implemented is not implemented"),
    ):
        factory.ResnapServiceFactory.get_service()


def test_should_return_local_service() -> None:
    # When
    service = factory.ResnapServiceFactory.get_service()

    # Then
    assert isinstance(service, LocalResnapService)


def test_should_return_same_instance_if_called_two_times() -> None:
    # When
    service_1 = factory.ResnapServiceFactory.get_service()
    service_2 = factory.ResnapServiceFactory.get_service()

    # Then
    assert service_1 == service_2


s3_secrets = {
    "access_key": "toto",
    "secret_key": "toto",
    "bucket_name": "toto",
}


@patch("importlib.util.find_spec", return_value=True)
@patch("resnap.services.boto_service.load_file", return_value=s3_secrets)
def test_should_return_boto_service_with_boto_extra(mock_find_spec: MagicMock, mock_load_file: MagicMock) -> None:
    # Given
    factory._resnap_config.save_to = "s3"

    # When
    service = factory.ResnapServiceFactory.get_service()

    # Then
    assert isinstance(service, BotoResnapService)


@patch("importlib.util.find_spec", return_value=None)
@patch("resnap.services.boto_service.load_file", return_value=s3_secrets)
def test_should_raise_without_boto_extra(mock_find_spec: MagicMock, mock_load_file: MagicMock) -> None:
    # Given
    factory._resnap_config.save_to = "s3"

    if "resnap.services.boto_service" in sys.modules:
        del sys.modules["resnap.boto"]
        del sys.modules["resnap.services.boto_service"]

    # When / Then
    with pytest.raises(
        ImportError,
        match=re.escape("Please install the boto extra to save to S3: `pip install resnap[boto]`"),
    ):
        factory.ResnapServiceFactory.get_service()



from unittest.mock import ANY, MagicMock

import pytest
from botocore.client import Config

from resnap.boto.config import S3Config
from resnap.boto.connection import get_s3_connection


def config_equal(config1: Config, config2: Config) -> bool:
    return (
        config1.signature_version == config2.signature_version
        and config1.s3 == config2.s3
        and config1.request_checksum_calculation == config2.request_checksum_calculation
        and config1.response_checksum_validation == config2.response_checksum_validation
    )


@pytest.fixture
def mock_s3_config() -> S3Config:
    return S3Config(
        access_key="test_access_key",
        secret_key="test_secret_key",
        bucket_name="test_bucket",
        region_name="test_region",
        endpoint_url="http://test-endpoint.com",
        signature_version="s3v4",
        force_path_style=True,
    )


@pytest.fixture
def mock_boto_client(mocker) -> MagicMock:
    return mocker.patch("resnap.boto.connection.boto3.client")


class TestS3Connection:
    def test_open_connection(self, monkeypatch, mock_boto_client: MagicMock, mock_s3_config: S3Config) -> None:
        # Given
        monkeypatch.setenv("REQUESTS_CA_BUNDLE", "/custom/path/to/ca.pem")
        expected_config = Config(
            signature_version=mock_s3_config.signature_version,
            s3={"addressing_style": "path"},
            request_checksum_calculation="when_required",
            response_checksum_validation="when_required",
        )

        # When
        connection = get_s3_connection(mock_s3_config)
        assert connection.is_open() is False
        connection.open()

        # Then
        mock_boto_client.assert_called_once_with(
            "s3",
            aws_access_key_id=mock_s3_config.access_key,
            aws_secret_access_key=mock_s3_config.secret_key,
            region_name=mock_s3_config.region_name,
            endpoint_url=mock_s3_config.endpoint_url,
            verify="/custom/path/to/ca.pem",
            config=ANY,
        )
        assert config_equal(expected_config, mock_boto_client.call_args.kwargs['config'])
        assert connection.is_open() is True

    def test_close_connection(self, mock_s3_config: S3Config) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)
        connection._connection = MagicMock()

        # When
        connection.close()

        # Then
        assert connection.is_open() is False

    def test_is_open(self, mock_s3_config: S3Config) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)
        connection._connection = MagicMock()
        assert connection.is_open() is True

        # When
        connection._connection = None

        # Then
        assert connection.is_open() is False

    def test_connection_property_with_already_connection(
        self, mock_boto_client: MagicMock, mock_s3_config: S3Config
    ) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)
        connection._connection = mock_boto_client

        # Then
        assert connection.connection == mock_boto_client

    def test_connection_property_without_connection(
        self, mock_boto_client: MagicMock, mock_s3_config: S3Config
    ) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)
        mock_client_instance = MagicMock()
        mock_boto_client.return_value = mock_client_instance

        # When
        con = connection.connection

        # Then
        mock_boto_client.assert_called_once()
        assert con == mock_client_instance

    def test_connection_property_setter(self, mock_s3_config: S3Config) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)

        # When / Then
        with pytest.raises(ValueError, match="The connection attribute is immutable."):
            connection.connection = MagicMock()

    def test_context_manager(self, mock_boto_client: MagicMock, mock_s3_config: S3Config) -> None:
        # Given
        mock_client_instance = MagicMock()
        mock_boto_client.return_value = mock_client_instance

        # When / Then
        with get_s3_connection(mock_s3_config) as conn:
            assert conn == mock_client_instance

    def test_context_manager_call_twice(self, mock_s3_config: S3Config) -> None:
        # Given
        connection = get_s3_connection(mock_s3_config)
        connection._connection = MagicMock()

        # When
        conn = connection._build_connection()

        # Then
        assert conn == connection._connection
        assert connection.is_open() is True



import asyncio
from datetime import datetime

import pandas as pd
from dependency_injector.wiring import Provide, inject

from resnap import ResnapError, async_resnap, resnap

exec_time = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


@resnap
def a_function(is_enable: bool) -> bool:
    return not is_enable


class BasicClass:
    @classmethod
    @resnap(output_format="txt")
    def get_param_value(cls, titi: str = "titi") -> str:
        return titi

    @classmethod
    @resnap
    def generate_dataframe(cls, titi: str = "titi") -> pd.DataFrame:
        return pd.DataFrame(
            {
                "A": [1, 2, 3],
                "B": [4, 5, 6],
                "C": [7, 8, 9],
            }
        )


class Pipeline:
    attribut_classe = "valeur"

    def __init__(self, titi: str = "valeur") -> None:
        self.attribut_instance = titi

    async def run(
        self,
        first_method_argument: pd.DataFrame | None = None,
        second_method_argument: str = "titi",
        third_method_argument: str = "titi",
        basic_class_get_param_value_argument: str = "titi",
        basic_class_generate_dataframe_argument: str = "titi",
        a_function_argument: bool = True,
    ) -> None:
        self._first_method(first_method_argument)
        await self._second_method(second_method_argument)
        BasicClass.get_param_value(basic_class_get_param_value_argument)
        BasicClass.generate_dataframe(basic_class_generate_dataframe_argument)
        a_function(a_function_argument)
        self._third_method(third_method_argument)

    @resnap(output_format="csv", output_folder=f"toto_{exec_time}")
    @inject
    def _first_method(self, df: pd.DataFrame | None = None, test=Provide["test"]) -> pd.DataFrame:
        if df is not None:
            return df
        return pd.DataFrame(
            {
                "A": [1, 2, 3],
                "B": [4, 5, 6],
                "C": [7, 8, 9],
            }
        )

    @async_resnap(output_format="txt")
    async def _second_method(self, titi: str = "titi") -> str:
        await asyncio.sleep(0.1)
        return titi

    @staticmethod
    @resnap(output_format="json", output_folder="test")
    def _third_method(titi: str = "titi") -> dict:
        if titi == "toto":
            raise ResnapError("toto is not allowed", data={"titi": titi})
        return {"tata": titi}



from abc import ABCMeta
from threading import Lock


class SingletonMeta(type):
    "This is a thread-safe implementation of Singleton."

    _instances: dict = {}

    _lock: Lock = Lock()
    "We now have a lock object that will be used to synchronize threads during first access to the Singleton."

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SingletonABCMeta(ABCMeta, SingletonMeta):
    "Allows combining the functionalities of both metaclasses without conflict."

    pass



import logging
import os

import toml

from .helpers.config import Config

CONFIG_SECTIONS: dict[str, str] = {
    "pyproject.toml": "resnap",
}

logger = logging.getLogger("resnap")


def get_config_data(file_path: str | None = None) -> Config:
    """
    Get the configuration data from the specified file or the default file.
    If no file is specified, it will look for a default configuration file.

    Args:
        file_path (str | None): Path to the configuration file. If None, it will look for a default file.

    Returns:
        Config: Configuration data.
    """
    if not file_path:
        file_path = get_config_file_path()
    if file_path.endswith(".toml"):
        config = toml.load(file_path)
        tools = config.get("tool", {})
        settings = tools.get(CONFIG_SECTIONS["pyproject.toml"], {"enabled": True})
        return Config(**settings)
    else:
        logger.warning(f"Unsupported file type: {file_path}.")
        return Config(enabled=True)


def get_config_file_path() -> str:
    """
    Get the path to the configuration file.

    Returns:
        str: Path to the configuration file.
    """
    exec_path: str = os.getcwd()
    for file_name in CONFIG_SECTIONS:
        if file_name in os.listdir(exec_path):
            return os.path.join(exec_path, file_name)
    return ""



from dataclasses import dataclass

from typing_extensions import Self

from resnap.helpers.config import Config, Services
from resnap.helpers.utils import TimeUnit


@dataclass
class ConfigBuilder:
    _enabled: bool = True
    _save_to: Services = Services.LOCAL
    _output_base_path: str = ""
    _secrets_file_name: str = ""
    _enable_remove_old_files: bool = False
    _max_history_files_length: int = 3
    _max_history_files_time_unit: TimeUnit = TimeUnit.HOUR

    @classmethod
    def a_config(cls) -> Self:
        return cls()

    def is_enabled(self, enabled: bool) -> Self:
        self._enabled = enabled
        return self

    def save_to(self, save_to: Services) -> Self:
        self._save_to = save_to
        return self

    def with_output_base_path(self, output_base_path: str) -> Self:
        self._output_base_path = output_base_path
        return self

    def with_secrets_file_name(self, secrets_file_name: str) -> Self:
        self._secrets_file_name = secrets_file_name
        return self

    def enable_remove_old_files(self, enable_remove_old_files: bool) -> Self:
        self._enable_remove_old_files = enable_remove_old_files
        return self

    def with_max_history_files_length(self, max_history_files_length: int) -> Self:
        self._max_history_files_length = max_history_files_length
        return self

    def with_max_history_files_time_unit(self, max_history_files_time_unit: TimeUnit) -> Self:
        self._max_history_files_time_unit = max_history_files_time_unit
        return self

    def build(self) -> Config:
        return Config(
            enabled=self._enabled,
            save_to=self._save_to,
            output_base_path=self._output_base_path,
            secrets_file_name=self._secrets_file_name,
            enable_remove_old_files=self._enable_remove_old_files,
            max_history_files_length=self._max_history_files_length,
            max_history_files_time_unit=self._max_history_files_time_unit,
        )



import json
import pickle
from datetime import datetime
from pathlib import Path
from typing import Any

import pandas as pd

from ..helpers.constants import EXT
from ..helpers.metadata import Metadata
from ..helpers.status import Status
from ..helpers.utils import calculate_datetime_from_now, get_datetime_from_filename
from .service import ResnapService


class LocalResnapService(ResnapService):
    def clear_old_saves(self) -> None:
        limit_time: datetime = calculate_datetime_from_now(
            self.config.max_history_files_length, self.config.max_history_files_time_unit,
        )
        folders: set[Path] = set()
        for f in Path(self.config.output_base_path).rglob("*"):
            if f.is_dir():
                folders.add(f)
                continue
            if f.is_file() and EXT in f.name:
                file_time = get_datetime_from_filename(f.name)
                if file_time < limit_time:
                    f.unlink()
        for folder in folders:
            contents = list(folder.iterdir())
            if not contents:
                folder.rmdir()

    def _create_folder(self, path: str, folder_name: str) -> None:
        folder_path = Path(path) / folder_name
        folder_path.mkdir(exist_ok=True)

    @staticmethod
    def _read_metadata(metadata_path: Path) -> Metadata:
        with open(metadata_path, "r") as json_file:
            data = json.load(json_file)
        return Metadata.from_dict(data)

    def get_success_metadata(self, func_name: str, output_folder: str) -> list[Metadata]:
        files: list[Path] = [
            Path(x) for x in Path(self._get_output_path(output_folder)).rglob(f"*{EXT}*")
            if x.is_file() and func_name in x.name and x.suffix == EXT
        ]

        if not files:
            return []

        metadata: list[Metadata] = [self._read_metadata(f) for f in sorted(files, reverse=True)]
        return [m for m in metadata if m.status == Status.SUCCESS]

    def _read_parquet_to_dataframe(self, file_path: str) -> pd.DataFrame:
        return pd.read_parquet(file_path)

    def _read_csv_to_dataframe(self, file_path: str) -> pd.DataFrame:
        return pd.read_csv(file_path, index_col=False)

    def _read_pickle(self, file_path: str) -> Any:
        with open(file_path, "rb") as f:
            return pickle.load(f)

    def _read_text(self, file_path: str) -> str:
        with open(file_path, "r") as f:
            return f.read()

    def _read_json(self, file_path: str) -> Any:
        with open(file_path, "r") as f:
            return json.load(f)

    @staticmethod
    def _save_dataframe_to_csv(result: pd.DataFrame, result_path: str) -> None:
        result.to_csv(result_path, index=False)

    @staticmethod
    def _save_dataframe_to_parquet(result: pd.DataFrame, result_path: str) -> None:
        result.to_parquet(result_path, compression="gzip")

    def _save_to_pickle(self, result: Any, result_path: str) -> None:
        with open(result_path, "wb") as f:
            pickle.dump(result, f)

    def _save_to_text(self, result: Any, result_path: str) -> None:
        with open(result_path, "w") as f:
            f.write(str(result))

    def _save_to_json(self, result: Any, result_path: str) -> None:
        with open(result_path, "w") as f:
            json.dump(result, f, indent=4)

    @staticmethod
    def _write_metadata(metadata_path: str, metadata: Metadata) -> None:
        with open(metadata_path, "w") as json_file:
            json.dump(metadata.to_dict(), json_file, indent=4)



"""The `version` module holds the version information for resnap."""

__all__ = "VERSION"

VERSION = "0.2.0"
"""The version of resnap."""


def version_short() -> str:
    """Return the `major.minor` part of resnap version.

    It returns '1.2' if resnap version is '1.2.6'.
    """
    return ".".join(VERSION.split(".")[:2])



from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from resnap.helpers.metadata import Metadata, MetadataSuccess
from resnap.helpers.status import Status
from resnap.helpers.utils import hash_arguments
from resnap.services.local_service import LocalResnapService
from tests.builders.config_builder import ConfigBuilder


@pytest.fixture
def mock_path_rglob(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("pathlib.Path.rglob")
    return mock


@pytest.fixture
def mock_path_mkdir(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("pathlib.Path.mkdir")
    return mock


def get_mock_path_file(name: str, is_file: bool) -> MagicMock:
    mock: MagicMock = MagicMock(
        spec=Path,
        is_dir=MagicMock(return_value=not is_file),
        is_file=MagicMock(return_value=is_file),
        unlink=MagicMock(),
    )
    mock.name = name
    if is_file:
        mock.suffix = f".{name.split('.')[-1]}"
    return mock


class TestLocalService:
    hashed_arguments: str = hash_arguments({"test": "toto"})

    @pytest.mark.parametrize(
        "file_path, is_file, is_deleted",
        [
            ("toto", False, False),
            ("test_2021-01-01T00-00-00.resnap", True, True),
            ("test_2021-01-01T00-00-00.resnap.pkl", True, True),
            ("toto/toto_2021-01-01T00-00-00.csv", True, False),
            (f"test_{datetime.now().isoformat().replace(':', '-')}.resnap", True, False),
        ],
    )
    def test_should_clear_old_saves(
        self,
        file_path: str,
        is_file: bool,
        is_deleted: bool,
        mock_path_rglob: MagicMock,
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        mock_file = get_mock_path_file(file_path, is_file)
        mock_path_rglob.return_value = [mock_file]

        # When
        service.clear_old_saves()

        # Then
        mock_path_rglob.assert_called_once()
        assert mock_file.unlink.call_count == int(is_deleted)

    def test_clear_old_saves_should_delete_empty_folder(
        self, mock_path_rglob: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        mock_file = get_mock_path_file("toto", False)
        mock_file.iterdir.return_value = []
        mock_path_rglob.return_value = [mock_file]

        # When
        service.clear_old_saves()

        # Then
        assert mock_file.unlink.call_count == 0
        mock_file.rmdir.assert_called_once()

    def test_clear_old_saves_should_not_delete_unempty_folder(
        self, mock_path_rglob: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        mock_file = get_mock_path_file("toto", False)
        mock_file.iterdir.return_value = ["test.csv"]
        mock_path_rglob.return_value = [mock_file]

        # When
        service.clear_old_saves()

        # Then
        assert mock_file.unlink.call_count == 0
        mock_file.rmdir.assert_not_called()

    def test_should_read_metadata(self) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        expected_metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            result_path="test_2021-01-01T00-00-00.resnap.pkl",
            result_type="str",
            hashed_arguments=self.hashed_arguments,
            extra_metadata={},
        )

        # When
        result = service._read_metadata(
            "tests/data/metadata/test-metadata_2021-01-01T00-00-00.resnap"
        )

        # Then
        assert isinstance(result, Metadata)
        assert result == expected_metadata

    def test_should_return_multiple_metadata(self, mock_path_rglob: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        mock_path_rglob.return_value = [
            Path("tests/data/metadata/test-metadata_2021-01-01T00-00-00.resnap"),
            Path("tests/data/metadata/test-metadata_2024-01-01T00-00-00.resnap"),
            get_mock_path_file("toto", False),
            get_mock_path_file("toto_2021-01-02T00-00-00.resnap", True),
            get_mock_path_file("test.csv", True),
        ]

        # When
        result = service.get_success_metadata("test", "")

        # Then
        assert result == [
            MetadataSuccess(
                status=Status.SUCCESS,
                event_time=datetime.fromisoformat("2024-01-01T00:00:00"),
                result_path="test_2024-01-01T00-00-00.resnap.pkl",
                result_type="str",
                hashed_arguments=hash_arguments({}),
                extra_metadata={},
            ),
            MetadataSuccess(
                status=Status.SUCCESS,
                event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
                result_path="test_2021-01-01T00-00-00.resnap.pkl",
                result_type="str",
                hashed_arguments=self.hashed_arguments,
                extra_metadata={},
            ),
        ]

    def test_should_return_none_if_not_metadata(
        self, mock_path_rglob: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        mock_path_rglob.return_value = []

        # When
        result = service.get_success_metadata("test", "toto")

        # Then
        assert len(result) == 0

    @patch("json.dump")
    @patch("builtins.open")
    def test_should_write_metadata(
        self, mock_open: MagicMock, mock_json_dump: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            result_path="test_2021-01-01T00-00-00.resnap.pkl",
            result_type="str",
            hashed_arguments=self.hashed_arguments,
        )

        # When
        service._write_metadata("test_2021-01-01T00-00-00.resnap", metadata)

        # Then
        mock_open.assert_called_once()
        mock_json_dump.assert_called_once_with(
            metadata.to_dict(), mock_open.return_value.__enter__(), indent=4
        )

    @patch("pandas.DataFrame.to_parquet")
    def test_should_save_dataframe_to_parquet(self, mock_to_parquet: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        result = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        result_path = "test.parquet"

        # When
        service._save_dataframe_to_parquet(result, result_path)

        # Then
        mock_to_parquet.assert_called_once_with(result_path, compression="gzip")

    @patch("pickle.dump")
    @patch("builtins.open")
    def test_should_save_to_pickle(
        self, mock_open: MagicMock, mock_pickle_dump: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        result = {"key": "value"}
        result_path = "test.pkl"

        # When
        service._save_to_pickle(result, result_path)

        # Then
        mock_open.assert_called_once_with(result_path, "wb")
        mock_pickle_dump.assert_called_once_with(
            result, mock_open.return_value.__enter__()
        )

    @patch("pandas.read_parquet")
    def test_should_read_parquet_to_dataframe(
        self, mock_read_parquet: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        file_path = "test.parquet"
        expected_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        mock_read_parquet.return_value = expected_df

        # When
        result = service._read_parquet_to_dataframe(file_path)

        # Then
        mock_read_parquet.assert_called_once_with(file_path)
        pd.testing.assert_frame_equal(result, expected_df)

    @patch("pickle.load")
    @patch("builtins.open")
    def test_should_read_pickle(
        self, mock_open: MagicMock, mock_pickle_load: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        file_path = "test.pkl"
        expected_data = {"key": "value"}
        mock_pickle_load.return_value = expected_data

        # When
        result = service._read_pickle(file_path)

        # Then
        mock_open.assert_called_once_with(file_path, "rb")
        mock_pickle_load.assert_called_once_with(mock_open.return_value.__enter__())
        assert result == expected_data

    @patch("json.load")
    @patch("builtins.open")
    def test_should_read_json(
        self, mock_open: MagicMock, mock_json_load: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        file_path = "test.json"
        expected_data = {"key": "value"}
        mock_json_load.return_value = expected_data

        # When
        result = service._read_json(file_path)

        # Then
        mock_open.assert_called_once_with(file_path, "r")
        mock_json_load.assert_called_once_with(mock_open.return_value.__enter__())
        assert result == expected_data

    @patch("builtins.open")
    def test_should_read_text(self, mock_open: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        file_path = "test.txt"
        expected_data = "test"
        mock_open.return_value.__enter__().read.return_value = expected_data

        # When
        result = service._read_text(file_path)

        # Then
        mock_open.assert_called_once_with(file_path, "r")
        assert result == expected_data

    @patch("pandas.DataFrame.to_csv")
    def test_should_save_dataframe_to_csv(self, mock_to_csv: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        result = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        result_path = "test.parquet"

        # When
        service._save_dataframe_to_csv(result, result_path)

        # Then
        mock_to_csv.assert_called_once_with(result_path, index=False)

    @patch("pandas.read_csv")
    def test_should_read_csv_to_dataframe(self, mock_read_csv: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        file_path = "test.csv"
        expected_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        mock_read_csv.return_value = expected_df

        # When
        result = service._read_csv_to_dataframe(file_path)

        # Then
        mock_read_csv.assert_called_once_with(file_path, index_col=False)
        pd.testing.assert_frame_equal(result, expected_df)

    @patch("builtins.open")
    def test_should_save_to_text(self, mock_open: MagicMock) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        result = "toto"
        result_path = "test.text"

        # When
        service._save_to_text(result, result_path)

        # Then
        mock_open.assert_called_once_with(result_path, "w")
        mock_open.return_value.__enter__().write.assert_called_once_with(result)

    @patch("json.dump")
    @patch("builtins.open")
    def test_should_save_to_json(
        self, mock_open: MagicMock, mock_json_dump: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(config=ConfigBuilder.a_config().build())
        result = {"key": "value"}
        result_path = "test.json"

        # When
        service._save_to_json(result, result_path)

        # Then
        mock_open.assert_called_once_with(result_path, "w")
        mock_json_dump.assert_called_once_with(
            result, mock_open.return_value.__enter__(), indent=4
        )

    @pytest.mark.parametrize(
        "path, folder_name",
        [
            ("toto", "folder"),
            ("", "folder"),
        ],
    )
    def test_should_create_folder(
        self, path: str, folder_name: str, mock_path_mkdir: MagicMock
    ) -> None:
        # Given
        service = LocalResnapService(ConfigBuilder.a_config().build())

        # When
        service._create_folder(path, folder_name)

        # Then
        assert mock_path_mkdir.call_count == 1



from unittest.mock import MagicMock

import pytest

from resnap.helpers.results_retriever import ResultsRetriever
from tests.builders.metadata_builder import MetadataSuccessBuilder


@pytest.fixture
def mock_service(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.helpers.results_retriever.ResnapService", return_value=MagicMock())
    return mock


@pytest.fixture
def mock_get_function_signature(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.helpers.results_retriever.get_function_signature")
    return mock


@pytest.fixture
def mock_hash_arguments(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.helpers.results_retriever.hash_arguments")
    return mock


class TestResultsRetriever:
    def test_should_not_return_results_if_recovery_is_disabled(self, mock_service: MagicMock) -> None:
        # Given
        retriever = ResultsRetriever(mock_service(), {"enable_recovery": False})

        # When
        result, is_recovery = retriever._get_saved_result()

        # Then
        assert result is None
        assert not is_recovery
        mock_service.return_value.read_result.assert_not_called()

    def test_should_not_return_results__if_no_metadata(self, mock_service: MagicMock) -> None:
        # Given
        retriever = ResultsRetriever(mock_service(), {"enable_recovery": True})
        mock_service.return_value.get_success_metadata.return_value = []

        # When
        result, is_recovery = retriever._get_saved_result()

        # Then
        assert result is None
        assert not is_recovery
        mock_service.return_value.get_success_metadata.assert_called_once()
        mock_service.return_value.read_result.assert_not_called()

    def test_should_not_check_arguments_if_disabled(self, mock_service: MagicMock) -> None:
        # Given
        retriever = ResultsRetriever(mock_service(), {"enable_recovery": True, "consider_args": False})
        metadata = MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 30}).build()
        mock_service.return_value.get_success_metadata.return_value = [metadata]
        mock_service.return_value.read_result.return_value = 30

        # When
        result, is_recovery = retriever._get_saved_result()

        # Then
        assert result == 30
        assert is_recovery is True
        mock_service.return_value.read_result.assert_called_once_with(metadata)
        mock_service.return_value.save_result.assert_not_called()

    def test_should_return_results_if_arguments_match(self, mock_service: MagicMock) -> None:
        # Given
        retriever = ResultsRetriever(mock_service(), {})
        expected_metadata = MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 30}).build()
        other_metadata = MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 666}).build()
        retriever.hashed_arguments = expected_metadata.hashed_arguments
        mock_service.return_value.get_success_metadata.return_value = [other_metadata, expected_metadata]
        mock_service.return_value.read_result.return_value = 30

        # When
        result, is_recovery = retriever._get_saved_result()

        # Then
        assert result == 30
        assert is_recovery is True
        mock_service.return_value.read_result.assert_called_once_with(expected_metadata)

    def test_should_not_return_results_if_arguments_not_match(self, mock_service: MagicMock) -> None:
        # Given
        retriever = ResultsRetriever(mock_service(), {})
        metadata = MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 30}).build()
        retriever.hashed_arguments = "toto"
        mock_service.return_value.get_success_metadata.return_value = [metadata]

        # When
        result, is_recovery = retriever._get_saved_result()

        # Then
        assert result is None
        assert is_recovery is False
        mock_service.return_value.read_result.assert_not_called()

    def test_should_return_saved_result(
        self,
        mock_service: MagicMock,
        mock_get_function_signature: MagicMock,
        mock_hash_arguments: MagicMock,
    ) -> None:
        # Given
        def func() -> None:
            pass

        retriever = ResultsRetriever(mock_service(), {"enable_recovery": False, "output_folder": "folder_name"})
        mock_get_function_signature.return_value = ("func_name", {"arg_1": 1, "arg_2": "value"})
        mock_hash_arguments.return_value = "hashed_arguments"

        # When
        result, is_recovery = retriever.get_results(func, (1,), {})

        # Then
        assert result is None
        assert not is_recovery
        mock_service.return_value.create_output_folder.assert_called_once_with("folder_name")
        mock_get_function_signature.assert_called_once_with(func, (1,), {}, [])
        mock_hash_arguments.assert_called_once_with({"arg_1": 1, "arg_2": "value"})



import io
from unittest.mock import MagicMock, call, patch

import pandas as pd
import pytest
from botocore.exceptions import ClientError

from resnap.boto.client import S3Client
from resnap.boto.config import S3Config


@pytest.fixture
def mock_s3_config() -> S3Config:
    return S3Config(
        access_key="test_access_key",
        secret_key="test_secret_key",
        bucket_name="test_bucket",
        region_name="test_region",
        endpoint_url="http://test-endpoint.com",
    )


@pytest.fixture
def mock_connection(mocker) -> MagicMock:
    mock = mocker.patch("resnap.boto.client.get_s3_connection")
    return mock


@pytest.fixture
def mock_open(mocker) -> MagicMock:
    mock = mocker.patch("builtins.open")
    return mock


@pytest.fixture
def mock_s3_client(mock_s3_config: S3Config) -> S3Client:
    return S3Client(config=mock_s3_config)


@pytest.fixture
def mock_dataframe_handler(mocker) -> MagicMock:
    mock = mocker.patch("resnap.boto.client.get_dataframe_handler")
    return mock


class TestS3Client:
    @pytest.mark.parametrize(
        "local_path_or_fileobj",
        [
            pytest.param(
                "test.txt",
                id="file_path",
            ),
            pytest.param(
                MagicMock(spec=io.FileIO),
                id="file_io",
            ),
            pytest.param(
                MagicMock(spec=io.BytesIO),
                id="bytes_io",
            ),
        ],
    )
    def test_should_upload_file(
        self,
        local_path_or_fileobj: str | io.FileIO | io.BytesIO,
        mock_s3_client: S3Client,
        mock_connection: MagicMock,
        mock_open: MagicMock,
    ) -> None:
        # Given
        if isinstance(local_path_or_fileobj, str):
            expected_fileobj = MagicMock()
            mock_open.return_value.__enter__.return_value = expected_fileobj
        else:
            expected_fileobj = local_path_or_fileobj

        # When
        mock_s3_client.upload_file(local_path_or_fileobj, "remote/test.txt")

        # Then
        mock_connection.return_value.__enter__.return_value.upload_fileobj.assert_called_once_with(
            expected_fileobj,
            mock_s3_client.config.bucket_name,
            "remote/test.txt",
        )

    @pytest.mark.parametrize(
        "local_path_or_fileobj",
        [
            pytest.param(
                "test.txt",
                id="file_path",
            ),
            pytest.param(
                MagicMock(spec=io.FileIO),
                id="file_io",
            ),
            pytest.param(
                MagicMock(spec=io.BytesIO),
                id="bytes_io",
            ),
        ],
    )
    def test_should_download_file(
        self,
        local_path_or_fileobj: str | io.FileIO | io.BytesIO,
        mock_connection: MagicMock,
        mock_s3_client: S3Client,
        mock_open: MagicMock,
    ) -> None:
        # Given
        if isinstance(local_path_or_fileobj, str):
            expected_fileobj = MagicMock()
            mock_open.return_value.__enter__.return_value = expected_fileobj
        else:
            expected_fileobj = local_path_or_fileobj

        # When
        mock_s3_client.download_file(local_path_or_fileobj, "remote/test.txt")

        # Then
        mock_connection.return_value.__enter__.return_value.download_fileobj.assert_called_once_with(
            mock_s3_client.config.bucket_name,
            "remote/test.txt",
            expected_fileobj,
        )

    @pytest.mark.parametrize("recursive", [True, False])
    @patch("resnap.boto.client.get_folders_and_files")
    def test_should_list_objects(
        self,
        mock_get_folders_and_files: MagicMock,
        recursive: bool,
        mock_s3_client: S3Client,
        mock_connection: MagicMock,
    ) -> None:
        # Given
        mock_connection.return_value.__enter__.return_value.get_paginator.return_value.paginate.return_value = [
            {"Contents": [{"Key": "file1.txt"}, {"Key": "file2.txt"}]}
        ]
        mock_get_folders_and_files.return_value = ([], ["file1.txt", "file2.txt"])
        expected_paginate_call = {
            "Bucket": mock_s3_client.config.bucket_name,
            "Prefix": "test/",
        }
        if not recursive:
            expected_paginate_call["Delimiter"] = "/"

        # When
        result = mock_s3_client.list_objects(remote_dir_path="test/", recursive=recursive)

        # Then
        assert result == ["file1.txt", "file2.txt"]
        mock_connection.return_value.__enter__.return_value.get_paginator.assert_called_once_with(
            "list_objects_v2"
        )
        mock_connection.return_value.__enter__.return_value.get_paginator.return_value.paginate.assert_called_once_with(
            **expected_paginate_call,
        )
        mock_get_folders_and_files.assert_called_once_with(
            "test/*", [{"Contents": [{"Key": "file1.txt"}, {"Key": "file2.txt"}]}]
        )

    def test_delete_object(self, mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # When
        mock_s3_client.delete_object("file1.txt")

        # Then
        mock_connection.return_value.__enter__.return_value.delete_object.assert_called_once_with(
            Bucket="test_bucket", Key="file1.txt"
        )

    @pytest.mark.parametrize(
        "keys",
        [
            pytest.param(["file1.txt", "file2.txt"], id="multiple_keys"),
            pytest.param(["file1.txt"], id="single_key"),
        ],
    )
    def test_should_delete_objects(self, keys: list[str], mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # Given
        expected_delete_call = {"Objects": [{"Key": key} for key in keys]}

        # When
        mock_s3_client.delete_objects(keys)

        # Then
        mock_connection.return_value.__enter__.return_value.delete_objects.assert_called_once_with(
            Bucket="test_bucket",
            Delete=expected_delete_call,
        )

    def test_should_delete_objects_empty(self, mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # When
        mock_s3_client.delete_objects([])

        # Then
        mock_connection.return_value.__enter__.return_value.delete_objects.assert_not_called()

    @pytest.mark.parametrize(
        "remote_path",
        [
            pytest.param("test.txt", id="object_exists"),
            pytest.param("folder/", id="folder_exist"),
            pytest.param("/", id="bucket_exist"),
        ],
    )
    def test_should_return_true_if_object_exists(
        self,
        remote_path: str,
        mock_s3_client: S3Client,
        mock_connection: MagicMock,
    ) -> None:
        # When
        result = mock_s3_client.object_exists(remote_path)

        # Then
        assert result is True
        if remote_path == "/":
            mock_connection.return_value.__enter__.return_value.get_bucket_acl.assert_called_once_with(
                Bucket="test_bucket"
            )
        elif remote_path.endswith("/"):
            mock_connection.return_value.__enter__.return_value.get_object.assert_called_once_with(
                Bucket="test_bucket", Key=remote_path
            )
        else:
            mock_connection.return_value.__enter__.return_value.head_object.assert_called_once_with(
                Bucket="test_bucket", Key=remote_path
            )

    def test_should_raise_if_object_not_exists(self, mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # Given
        mock_connection.return_value.__enter__.return_value.head_object.side_effect = ClientError(
            {"Error": {"Code": "404"}}, "HeadObject"
        )

        # When
        result = mock_s3_client.object_exists("file1.txt")

        # Then
        assert result is False

    @pytest.mark.parametrize(
        "path",
        [
            pytest.param("test/dir", id="dir_path"),
            pytest.param("test/dir/", id="dir_path_with_slash"),
        ],
    )
    def test_should_create_dir(self, path: str, mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # When
        mock_s3_client.mkdir(path)

        # Then
        mock_connection.return_value.__enter__.return_value.upload_fileobj.assert_called_once()

    def test_should_return_df_from_file(
        self,
        mock_s3_client: S3Client,
        mock_connection: MagicMock,
        mock_dataframe_handler: MagicMock,
    ) -> None:
        # Given
        mock_dataframe_handler.return_value.read_df.return_value = MagicMock()
        mock_connection.return_value.__enter__.return_value.get_object.return_value = {
            "Body": MagicMock(read=MagicMock(return_value=b"col1,col2\n1,2\n3,4"))
        }

        # When
        with patch.object(S3Client, "object_exists", return_value=True):
            result = mock_s3_client.get_df_from_file("test.csv", "csv", compression="gzip", nrows=10)

        # Then
        assert result == mock_dataframe_handler.return_value.read_df.return_value
        mock_connection.return_value.__enter__.return_value.get_object.assert_called_once_with(
            Bucket="test_bucket", Key="test.csv"
        )
        mock_dataframe_handler.return_value.read_df.assert_called_once_with(
            mock_connection.return_value.__enter__.return_value.get_object.return_value["Body"],
            compression="gzip",
            nrows=10,
        )

    def test_should_raise_if_file_not_exists(self, mock_s3_client: S3Client, mock_dataframe_handler: MagicMock) -> None:
        # When / Then
        with pytest.raises(FileNotFoundError, match="File test.csv not exists in bucket test_bucket"):
            with patch.object(S3Client, "object_exists", return_value=False):
                mock_s3_client.get_df_from_file("test.csv", "csv")

    @pytest.mark.parametrize(
        "compression, file_format, expected_compression",
        [
            pytest.param(None, "csv", None, id="no_compression"),
            pytest.param("gzip", "csv", "gzip", id="gzip_compression"),
            pytest.param(None, "parquet", "snappy", id="snappy_compression"),
        ],
    )
    def test_should_push_df_to_file(
        self,
        compression: str | None,
        file_format: str,
        expected_compression: str | None,
        mock_s3_client: S3Client,
        mock_dataframe_handler: MagicMock,
    ) -> None:
        # Given
        df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        expected_write_df_return = b"col1,col2\n1,2\n3,4"
        mock_dataframe_handler.return_value.write_df.return_value = expected_write_df_return

        # When / Then
        with patch.object(S3Client, "upload_file") as mock_upload_file:
            if compression:
                mock_s3_client.push_df_to_file(df, "test.csv", compression=compression, file_format=file_format)
            else:
                mock_s3_client.push_df_to_file(df, "test.csv", file_format=file_format)
            mock_dataframe_handler.return_value.write_df.assert_called_once_with(
                df, compression=expected_compression
            )
            mock_upload_file.assert_called_once_with(expected_write_df_return, "test.csv")

    def test_should_rmdir(self, mock_s3_client: S3Client, mock_connection: MagicMock) -> None:
        # Given
        mock_s3_client.list_folders_and_files = MagicMock(return_value=(["folder1/", "folder2/"], ["file1.txt"]))
        mock_s3_client.delete_objects = MagicMock()

        # When
        mock_s3_client.rmdir("test/")

        # Then
        assert mock_s3_client.delete_objects.call_count == 2
        calls = [
            call(["file1.txt"]),
            call(["folder1/", "folder2/"]),
        ]
        mock_s3_client.delete_objects.assert_has_calls(calls)



import argparse


def main(run_local: bool = True, run_metadata: bool = True, run_boto: bool = True) -> None:
    """
    Main function to run functional tests.
    Args:
        run_local (bool): Flag to run local tests.
        run_metadata (bool): Flag to run add_metadata tests.
        run_boto (bool): Flag to run boto tests.
    """
    print()
    print("#### LET'S START FUNCTIONAL TESTS ####")

    if run_local:
        from test_local_service import run as test_local_service
        test_local_service()

    if run_metadata:
        from test_add_metadata import run as test_add_metadata
        test_add_metadata()

    if run_boto:
        from test_boto_service import run as test_boto_service
        test_boto_service()

    print("#### FUNCTIONAL TESTS ALL PASSED ####")
    print("#### BRAVOOO ####")
    print()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run functional tests.")

    parser.add_argument("--local", action="store_true", help="Run local tests")
    parser.add_argument("--no-local", dest="local", action="store_false", help="Do not run local tests")
    parser.set_defaults(local=True)

    parser.add_argument("--metadata", action="store_true", help="Run add_metadata tests")
    parser.add_argument("--no-metadata", dest="metadata", action="store_false", help="Do not run add_metadata tests")
    parser.set_defaults(metadata=True)

    parser.add_argument("--boto", action="store_true", help="Run boto tests")
    parser.add_argument("--no-boto", dest="boto", action="store_false", help="Do not run boto tests")
    parser.set_defaults(boto=True)

    args = parser.parse_args()
    main(args.local, args.metadata, args.boto)



import importlib.util

if importlib.util.find_spec("boto3") is None:
    raise ImportError("Please install the boto extra to save to S3: `pip install resnap[boto]`")

from .client import S3Client
from .config import S3Config

__all__ = [
    "S3Client",
    "S3Config",
]
